﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using JetBrains.Annotations;
using System;
using System.Collections.Concurrent.Partitioners;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading.Tasks;
using TamiFlu.Interfaces;

namespace TamiFlu.Core
{
    [TypeConverter(typeof (ExpandableObjectConverter)), Plugin("Parallel", "1.1", "Stefan Band")]
    internal sealed class ParallelContext : IParallelContext
    {
        #region Properties

        [NotNull]
        private ParallelOptions Options => new ParallelOptions
                                           {
                                            #if DEBUG
                                                MaxDegreeOfParallelism = 1
                                            #else
                                               MaxDegreeOfParallelism = NumThreads
                                            #endif
                                           };

        #endregion

        #region Implementation of IParallelContext

        /// <inheritdoc />
        [Browsable(true)]
        public int NumThreads { get; set; } = Environment.ProcessorCount;

        /// <inheritdoc />
        public void ForEachRange(int startIndex, int endIndex, Action<int, int> action)
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            if (endIndex > startIndex)
            {
                System.Threading.Algorithms.ParallelAlgorithms.ForRange(startIndex, endIndex, Options, action);
            }
        }

        /// <inheritdoc />
        public void ForEach<T, TList>(TList source, Action<T> action) where TList : class, IReadOnlyList<T>
        {
            if (source is null)
            {
                throw new ArgumentNullException(nameof(source));
            }

            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            ForEachRange(0, source.Count, (startIndex, endIndex) =>
            {
                for (var i = startIndex; i < endIndex; ++i)
                {
                    action(source[i]);
                }
            });
        }

        /// <inheritdoc />
        public void ForEach<T>(IEnumerable<T> source, Action<T> action)
        {
            if (source is null)
            {
                throw new ArgumentNullException(nameof(source));
            }

            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            Parallel.ForEach(source, Options, action);
        }

        /// <inheritdoc />
        public void ForEach<T>(IEnumerable<T> source, int minChunkSize, int maxChunkSize, Action<T> action)
        {
            if (source is null)
            {
                throw new ArgumentNullException(nameof(source));
            }

            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            Parallel.ForEach(ChunkPartitioner.Create(source, minChunkSize, maxChunkSize), Options, action);
        }

        /// <inheritdoc />
        public TResult ReduceRange<TResult>(int startIndex, int endIndex, Func<int, TResult> action, Func<TResult, TResult, TResult> reduction, TResult initValue) where TResult : struct
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            if (reduction is null)
            {
                throw new ArgumentNullException(nameof(reduction));
            }

            return System.Threading.Algorithms.ParallelAlgorithms.Reduce(startIndex, endIndex, Options, action, initValue, reduction);
        }

        /// <inheritdoc />
        public TResult Reduce<TResult, T, TList>(TList source, Func<T, TResult> action, Func<TResult, TResult, TResult> reduction, TResult initValue) where TList : class, IReadOnlyList<T> where TResult : struct
        {
            if (source is null)
            {
                throw new ArgumentNullException(nameof(source));
            }

            return ReduceRange(0, source.Count, i => action(source[i]), reduction, initValue);
        }

        /// <inheritdoc />
        public TResult CombineRange<TResult>(int startIndex, int endIndex, Action<int, TResult> action, Func<TResult, TResult, TResult> reduction, Func<TResult> createInitValue) where TResult : class
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            if (reduction is null)
            {
                throw new ArgumentNullException(nameof(reduction));
            }

            using (var combinable = new ParallelThreadSpecific<TResult>(createInitValue))
            {
                ForEachRange(startIndex, endIndex, (firstIndex, lastIndex) =>
                {
                    // ReSharper disable once AccessToDisposedClosure
                    var value = combinable.ThreadLocal;

                    for (var i = firstIndex; i < lastIndex; ++i)
                    {
                        action(i, value);
                    }
                });

                return combinable.Reduce(reduction);
            }
        }

        /// <inheritdoc />
        public TResult Combine<TResult, T, TList>(TList source, Action<T, TResult> action, Func<TResult, TResult, TResult> reduction, Func<TResult> createInitValue) where TList : class, IReadOnlyList<T> where TResult : class
        {
            return CombineRange(0, source.Count, (index, value) => action(source[index], value), reduction, createInitValue);
        }

        #endregion
    }
}
