﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Threading;
using JetBrains.Annotations;

namespace TamiFlu.Core
{
    internal sealed class ParallelThreadSpecific<T> : IDisposable where T : class
    {
        #region Nested Types : Node

        private sealed class Node : IDisposable
        {
            public readonly int Key;
            public readonly T Value;
            public Node Next;

            public Node(T value, int key)
            {
                Value = value;
                Key = key;
                Next = null;
            }

            /// <inheritdoc />
            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            /// <inheritdoc />
            ~Node()
            {
                Dispose(false);
            }

            private void Dispose(bool disposing)
            {
                if (Value is IDisposable disposable)
                {
                    disposable.Dispose();
                }
            }
        }

        #endregion

        #region Properties

        [NotNull]
        public T ThreadLocal => GetThreadLocal(out var exists);

        #endregion

        #region Fields

        [NotNull] private readonly Func<T> m_CreateInstance;
        [NotNull] private readonly Node[] m_Buckets;

        #endregion

        #region Constructor

        public ParallelThreadSpecific([NotNull] Func<T> createInstance)
        {
            m_CreateInstance = createInstance ?? throw new ArgumentNullException(nameof(createInstance));
            m_Buckets = new Node[Environment.ProcessorCount];
        }

        /// <inheritdoc />
        ~ParallelThreadSpecific()
        {
            Dispose(false);
        }

        #endregion

        #region Methods

        /// <inheritdoc />
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            Clear();
        }

        private static int GetCurrentThreadID()
        {
            return Thread.CurrentThread.ManagedThreadId;
        }

        [NotNull]
        public T GetThreadLocal(out bool exists)
        {
            return GetThreadLocalNode(out exists).Value;
        }

        [NotNull]
        private Node GetThreadLocalNode(out bool exists)
        {
            var key = GetCurrentThreadID();
            var index = key.GetHashCode() % m_Buckets.Length;

            var node = m_Buckets[index];
            while (!(node is null))
            {
                if (node.Key == key)
                {
                    exists = true;
                    return node;
                }

                node = node.Next;
            }

            // node not found, i.e. create a new one
            exists = false;
            node = new Node(m_CreateInstance(), key);
            var firstNode = m_Buckets[index];
            while (true)
            {
                node.Next = firstNode;
                // try to make node the new firstNode
                var oldNode = Interlocked.CompareExchange(ref m_Buckets[index], node, firstNode);
                if (ReferenceEquals(oldNode, firstNode))
                {
                    return node;
                }
            }
        }

        public void Clear()
        {
            for (var index = 0; index < m_Buckets.Length; ++index)
            {
                Node node;
                while (!((node = Interlocked.CompareExchange(ref m_Buckets[index], null, null)) is null))
                {
                    do
                    {
                        node.Dispose();
                        node = node.Next;
                    } while (!(node is null));
                }
            }
        }

        [NotNull]
        public T Reduce([NotNull] Func<T, T, T> reduction)
        {
            var result = m_CreateInstance();

            foreach (var bucket in m_Buckets)
            {
                for (var node = bucket; !(node is null); node = node.Next)
                {
                    result = reduction(result, node.Value);
                }
                
            }

            return result;
        }

        #endregion
    }
}
