﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System.ComponentModel;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.Kernel;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Forces;

namespace TamiFlu.Core.Simulation.Forces
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("Surface Tension", "1.0", "Stefan Band"), UsedImplicitly]
    internal sealed class SurfaceTensionForce : ISurfaceTensionForce
    {
        #region Implementation of ISurfaceTensionForce

        /// <inheritdoc />
        public void ApplyToFluidParticles(IParticleContext particleContext)
        {
            var kernel = particleContext.Kernel;
            var fluidParticles = particleContext.FluidParticles;
            var normals = new Vector[fluidParticles.Length];

            particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
            {
                var normal = Vector.Zero;

                foreach (var ffIndex in f.FluidNeighbors)
                {
                    ref readonly var ff = ref fluidParticles[ffIndex];
                    normal += ff.Volume * kernel.ComputeGradient(f.Position, ff.Position);
                }

                normals[fIndex] = kernel.SupportLength * normal;
            });

            var cohesionKernel = new CohesionKernel();
            particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
            {
                if (f.Properties.SurfaceTension > 0.0f)
                {
                    var acceleration = Vector.Zero;

                    foreach (var ffIndex in f.FluidNeighbors)
                    {
                        ref readonly var ff = ref fluidParticles[ffIndex];
                        var cohesionAcceleration = ff.Properties.Mass * cohesionKernel.ComputeGradient(f.Position, ff.Position);
                        var curvatureAcceleration = normals[fIndex] - normals[ffIndex];
                        var k = 2f * f.Volume * ff.Volume / (f.Volume + ff.Volume);
                        acceleration -= k * (cohesionAcceleration + curvatureAcceleration);
                    }

                    f.Acceleration += f.Properties.SurfaceTension * acceleration;
                }
            });
        }

        #endregion
    }
}
