﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.ComponentModel;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Kernel;

namespace TamiFlu.Core.Simulation.Kernel
{
    [Plugin("Cohesion Kernel", "1.0", "Stefan Band", false)]
    internal sealed class CohesionKernel : IKernel
    {
        #region Constants

        private const float DefaultSupportFactor = 2f;

        #endregion

        #region Fields

        private float m_Alpha;
        private float m_SupportFactor;

        #endregion

        #region Constructor

        public CohesionKernel()
        {
            SupportFactor = DefaultSupportFactor;
        }

        #endregion

        #region Implementation of IKernel

        /// <inheritdoc />
        [Browsable(true), DefaultValue(DefaultSupportFactor)]
        public float SupportFactor
        {
            get { return m_SupportFactor; }
            set
            {
                m_SupportFactor = value;

                var particleSpacing = GlobalConstants.Instance.Spacing;
                SupportLength = particleSpacing * value;

                #pragma warning disable 162
                switch (Vector.Dimensions)
                {
                    case 2:
                        // TODO compute normalization factor for 2d simulations

                    case 3:
                        m_Alpha = (float) (32.0 / (Math.PI * Math.Pow(SupportLength, 9.0)));
                        break;

                    default:
                        throw new ArgumentOutOfRangeException();
                }
                #pragma warning restore 162
            }
        }

        /// <inheritdoc />
        public float SupportLength { get; private set; }

        /// <inheritdoc />
        public float ComputeValue(in Vector position1, in Vector position2)
        {
            throw new NotSupportedException();
        }

        /// <inheritdoc />
        public Vector ComputeGradient(in Vector position1, in Vector position2)
        {
            var distanceVector = position1 - position2;
            var distance = Vector.Length(in distanceVector);
            if (distance <= 0f || distance >= SupportLength)
            {
                return Vector.Zero;
            }

            var supportLengthMinusDistance = SupportLength - distance;
            var gradientFactor = (supportLengthMinusDistance * supportLengthMinusDistance * supportLengthMinusDistance) * (distance * distance * distance);
            if (distance > SupportLength / 2f)
            {
                gradientFactor *= 2f;
                var supportLength2 = SupportLength * SupportLength;
                gradientFactor -= supportLength2 * supportLength2 * supportLength2 / 64f;
            }

            return distanceVector * (m_Alpha / distance * gradientFactor);
        }

        /// <inheritdoc />
        public bool IsInRange(in Vector position1, in Vector position2)
        {
            return Vector.DistanceSquared(in position1, in position2) < SupportLength * SupportLength;
        }
        
        #endregion
    }
}
