﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using JetBrains.Annotations;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Kernel;

namespace TamiFlu.Core.Simulation.Kernel
{
    [Plugin("Cubic Spline", "1.1", "Stefan Band"), UsedImplicitly]
    internal sealed class CubicSplineKernel : IKernel
    {
        #region Constants

        private const float DefaultSupportFactor = 2f;

        #endregion

        #region Fields

        private float m_Alpha;
        private float m_SupportFactor;

        #endregion

        #region Constructor

        public CubicSplineKernel()
        {
            SupportFactor = DefaultSupportFactor;
        }

        #endregion

        #region Implementation of IKernel

        /// <inheritdoc />
        [Browsable(true), DefaultValue(DefaultSupportFactor)]
        public float SupportFactor
        {
            get => m_SupportFactor;
            set
            {
                m_SupportFactor = value;

                var particleSpacing = GlobalConstants.Instance.Spacing;
                SupportLength = particleSpacing * value;

                #pragma warning disable 162
                switch (Vector.Dimensions)
                {
                    case 2:
                        m_Alpha = (float) (5.0 / (14.0 * Math.PI * particleSpacing * particleSpacing));
                        break;

                    case 3:
                        m_Alpha = (float) (1.0 / (4.0 * Math.PI * particleSpacing * particleSpacing * particleSpacing));
                        break;

                    default:
                        throw new ArgumentOutOfRangeException();
                }
                #pragma warning restore 162
            }
        }

        /// <inheritdoc />
        [Browsable(false)]
        public float SupportLength { get; private set; }

        /// <inheritdoc />
        public float ComputeValue(in Vector position1, in Vector position2)
        {
            var q = ComputeQ(Vector.Distance(in position1, in position2));
            var oneMinusQ = Math.Max(1f - q, 0f);
            var twoMinusQ = Math.Max(2f - q, 0f);
            return m_Alpha * (twoMinusQ*twoMinusQ*twoMinusQ - 4f*oneMinusQ*oneMinusQ*oneMinusQ);
        }

        /// <inheritdoc />
        public Vector ComputeGradient(in Vector position1, in Vector position2)
        {
            var distanceVector = position1 - position2;
            var distance = Vector.Length(in distanceVector);
            if (distance <= 0.0f)
            {
                return Vector.Zero;
            }

            var q = ComputeQ(distance);
            var oneMinusQ = Math.Max(1f - q, 0f);
            var twoMinusQ = Math.Max(2f - q, 0f);
            var gradientFactor = -3f*twoMinusQ*twoMinusQ + 12f*oneMinusQ*oneMinusQ;
            return distanceVector * (m_Alpha / (SupportLength * distance) * gradientFactor);
        }

        /// <inheritdoc />
        public bool IsInRange(in Vector position1, in Vector position2)
        {
            return Vector.DistanceSquared(in position1, in position2) < (SupportLength * SupportLength);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        private float ComputeQ(float distance)
        {
            // distance is in the interval [0, SupportLength], i.e. q is in the interval [0, 1]
            var q = distance / SupportLength;

            // the returned value should be in the interval [0, DefaultSupportFactor], i.e.
            // we have to scale q by DefaultSupportFactor 
            return DefaultSupportFactor * q;
        }

        #endregion
    }
}
