﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using JetBrains.Annotations;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Neighborhood;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.Simulation.Neighborhood
{
    [Plugin("Naive O(N²)", "1.1", "Stefan Band"), UsedImplicitly]
    internal sealed class NaiveParticleNeighborhood : IParticleNeighborhood
    {
        #region Methods

        internal static (NeighborSet, NeighborSet) ClearNeighborSets<T>(ref T particle) where T : struct, IParticle
        {
            var fluidNeighbors = particle.FluidNeighbors as NeighborSet;
            if (fluidNeighbors is null)
            {
                particle.FluidNeighbors = fluidNeighbors = new NeighborSet();
            }
            else
            {
                fluidNeighbors.Clear();
            }

            var boundaryNeighbors = particle.BoundaryNeighbors as NeighborSet;
            if (boundaryNeighbors is null)
            {
                particle.BoundaryNeighbors = boundaryNeighbors = new NeighborSet();
            }
            else
            {
                boundaryNeighbors.Clear();
            }

            return (fluidNeighbors, boundaryNeighbors);
        }

        #endregion

        #region Implementation of IParticleNeighborhood

        /// <inheritdoc />
        public bool HasNeighbors { get; private set; }

        /// <inheritdoc />
        public void SearchNeighbors(IParticleContext particleContext)
        {
            HasNeighbors = false;

            var kernel = particleContext.Kernel;
            var fluidParticles = particleContext.FluidParticles;
            var boundaryParticles = particleContext.BoundaryParticles;

            particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
            {
                var (fluidNeighbors, boundaryNeighbors) = ClearNeighborSets(ref f);

                // fluid particles less than fIndex
                for (var ffIndex = ParticleIndex.Zero; ffIndex < fIndex; ++ffIndex)
                {
                    if (kernel.IsInRange(f.Position, fluidParticles[ffIndex].Position))
                    {
                        fluidNeighbors.Add(ffIndex);
                    }
                }

                // fluid particles greater than fIndex
                for (var ffIndex = fIndex + ParticleIndex.One; ffIndex < fluidParticles.Length; ++ffIndex)
                {
                    if (kernel.IsInRange(f.Position, fluidParticles[ffIndex].Position))
                    {
                        fluidNeighbors.Add(ffIndex);
                    }
                }

                // boundary particles
                for (var fbIndex = ParticleIndex.Zero; fbIndex < boundaryParticles.Length; ++fbIndex)
                {
                    if (kernel.IsInRange(f.Position, boundaryParticles[fbIndex].Position))
                    {
                        boundaryNeighbors.Add(fbIndex);
                    }
                }
            });

            particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b, in ParticleIndex bIndex) =>
            {
                var (fluidNeighbors, boundaryNeighbors) = ClearNeighborSets(ref b);

                // fluid particles
                for (var bfIndex = ParticleIndex.Zero; bfIndex < fluidParticles.Length; ++bfIndex)
                {
                    if (kernel.IsInRange(b.Position, fluidParticles[bfIndex].Position))
                    {
                        fluidNeighbors.Add(bfIndex);
                    }
                }

                // boundary particles less than bIndex
                for (var bbIndex = ParticleIndex.Zero; bbIndex < bIndex; ++bbIndex)
                {
                    if (kernel.IsInRange(b.Position, boundaryParticles[bbIndex].Position))
                    {
                        boundaryNeighbors.Add(bbIndex);
                    }
                }

                // boundary particles greater than fIndex
                for (var bbIndex = bIndex + ParticleIndex.One; bbIndex < boundaryParticles.Length; ++bbIndex)
                {
                    if (kernel.IsInRange(b.Position, boundaryParticles[bbIndex].Position))
                    {
                        boundaryNeighbors.Add(bbIndex);
                    }
                }
            });

            HasNeighbors = true;
        }

        /// <inheritdoc />
        public void SearchNeighbors(IParticleContext particleContext, Vector position, FluidParticleAction fluidParticleAction, BoundaryParticleAction boundaryParticleAction)
        {
            var kernel = particleContext.Kernel;

            particleContext.ForEachFluidParticle((ref FluidParticle f) =>
            {
                if (kernel.IsInRange(f.Position, position))
                {
                    fluidParticleAction(ref f);
                }
            });

            particleContext.ForEachBoundaryParticle((ref BoundaryParticle b) =>
            {
                if (kernel.IsInRange(b.Position, position))
                {
                    boundaryParticleAction(ref b);
                }
            });
        }

        #endregion
    }
}
