﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading.Algorithms;
using JetBrains.Annotations;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Neighborhood;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.Simulation.Neighborhood
{
    internal struct CellIndex : IEquatable<CellIndex>, IComparable<CellIndex>
    {
        #region Fields

        public readonly long Value;

        public static readonly CellIndex Zero = new CellIndex(0L);
        public static readonly CellIndex One = new CellIndex(1L);
        public static readonly CellIndex Invalid = new CellIndex(long.MaxValue);

        #endregion

        #region Constructor

        public CellIndex(long value)
        {
            Value = value;
        }

        #endregion

        #region Methods

        /// <inheritdoc />
        public bool Equals(CellIndex other)
        {
            return Value == other.Value;
        }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
            return obj is CellIndex && Equals((CellIndex) obj);
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            return Value.GetHashCode();
        }

        /// <inheritdoc />
        public int CompareTo(CellIndex other)
        {
            return Value.CompareTo(other.Value);
        }

        public static bool operator ==(CellIndex lhs, CellIndex rhs)
        {
            return lhs.Value == rhs.Value;
        }

        public static bool operator !=(CellIndex lhs, CellIndex rhs)
        {
            return lhs.Value != rhs.Value;
        }

        public static CellIndex operator +(CellIndex lhs, CellIndex rhs)
        {
            return new CellIndex(lhs.Value + rhs.Value);
        }

        public static CellIndex operator -(CellIndex lhs, CellIndex rhs)
        {
            return new CellIndex(lhs.Value - rhs.Value);
        }

        #endregion
    }

    internal interface ICellIndexComputer
    {
        #region Methods

        void Setup(in BoundingBox boundingBox, float cellSize);

        CellIndex ComputeCellIndex(in Vector position);

        void ForEachNeighborCellIndex(in CellIndex cellIndex, [NotNull] Action<CellIndex> action);

        #endregion
    }

    internal abstract class ABaseCellIndex : ICellIndexComputer
    {
        #region Constants

        protected const int NumberOfLayers = 1;

        #endregion

        #region Fields

        protected BoundingBox BoundingBox;
        protected int NumberOfCellsX;
        protected int NumberOfCellsY;
        protected float CellSize;

        #endregion

        #region Implementation of ICellIndexComputer

        /// <inheritdoc />
        public void Setup(in BoundingBox boundingBox, float cellSize)
        {
            BoundingBox = boundingBox;
            var size = boundingBox.Size;
            var numberOfCells = Vector.Ceiling(size / cellSize);
            NumberOfCellsX = (int) numberOfCells.X + 1;
            NumberOfCellsY = (int) numberOfCells.Y + 1;
            CellSize = cellSize;
        }

        /// <inheritdoc />
        public CellIndex ComputeCellIndex(in Vector position)
        {
            var cellCoordinate = Vector.Floor(Vector.Max(Vector.Zero, (position - BoundingBox.Minimum) / CellSize));
            return ComputeCellIndex2D((int) cellCoordinate.X, (int) cellCoordinate.Y);
        }

        protected abstract CellIndex ComputeCellIndex2D(int cellCoordinateX, int cellCoordinateY);

        protected abstract (int X, int Y) ComputeCellCoordinate2D(in CellIndex cellIndex);

        /// <inheritdoc />
        public void ForEachNeighborCellIndex(in CellIndex cellIndex, Action<CellIndex> action)
        {
            var cellCoordinate2D = ComputeCellCoordinate2D(in cellIndex);
            for (var dy = -NumberOfLayers; dy <= NumberOfLayers; ++dy)
            {
                for (var dx = -NumberOfLayers; dx <= NumberOfLayers; ++dx)
                {
                    action(ComputeCellIndex2D(cellCoordinate2D.X + dx, cellCoordinate2D.Y + dy));
                }
            }
        }
        #endregion
    }

    [Plugin("SortXYZ Indexer", "1.1", "Stefan Band"), UsedImplicitly]
    internal sealed class DefaultCellIndex : ABaseCellIndex
    {
        #region Constants

        private const bool UseDynamicIndexBits = false;
        private const int StaticIndexBits = (8 * sizeof(long)) / Vector.Dimensions;

        #endregion

        #region Static Methods

        /// <summary>
        /// Extract contiguous bits from value. Extract the number of bits specified by length, starting at the bit specified by start.
        /// </summary>
        private static long ExtractBits(long value, int start, int length)
        {
            unchecked
            {
                return (value >> start) & ((1L << length) - 1L);
            }
        }

        #endregion

        #region Implementation of ICellIndexComputer

        /// <inheritdoc />
        protected override CellIndex ComputeCellIndex2D(int cellCoordinateX, int cellCoordinateY)
        {
            unchecked
            {
                // ReSharper disable once ConditionIsAlwaysTrueOrFalse
                if (UseDynamicIndexBits)
                {
                    var numberOfCellsX = (long) NumberOfCellsX;
                    return new CellIndex(cellCoordinateX + cellCoordinateY * numberOfCellsX);
                }
                else
                {
                    return new CellIndex((long) cellCoordinateX | ((long) cellCoordinateY << StaticIndexBits));
                }
            }
        }

        /// <inheritdoc />
        protected override (int X, int Y) ComputeCellCoordinate2D(in CellIndex cellIndex)
        {
            unchecked
            {
                // ReSharper disable once ConditionIsAlwaysTrueOrFalse
                if (UseDynamicIndexBits)
                {
                    var numberOfCellsX = (long) NumberOfCellsX;
                    return ((int) (cellIndex.Value % numberOfCellsX), (int) (cellIndex.Value / numberOfCellsX));
                }
                else
                {
                    return ((int) ExtractBits(cellIndex.Value, 0, StaticIndexBits),
                            (int) ExtractBits(cellIndex.Value, StaticIndexBits, StaticIndexBits));
                }
            }
        }
        
        #endregion
    }

    [Plugin("Z-Indexer", "1.0", "Stefan Band"), UsedImplicitly]
    internal sealed class MortonCellIndex : ABaseCellIndex
    {
        #region Implementation of ICellIndexComputer

        /// <inheritdoc />
        protected override CellIndex ComputeCellIndex2D(int cellCoordinateX, int cellCoordinateY)
        {
            return new CellIndex(unchecked ((long) new Morton2D(cellCoordinateX, cellCoordinateY).Key));
        }

        /// <inheritdoc />
        protected override (int X, int Y) ComputeCellCoordinate2D(in CellIndex cellIndex)
        {
            return new Morton2D(unchecked((ulong) cellIndex.Value)).Decode();
        }

        #endregion
    }

    [TypeConverter(typeof (ExpandableObjectConverter)), Plugin("Uniform Grid", "1.01", "Stefan Band"), UsedImplicitly]
    internal sealed class UniformGridParticleNeighborhood : IParticleNeighborhood
    {
        #region Fields

        private ConcurrentDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>> m_FluidParticleGrid;
        private ConcurrentDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>> m_BoundaryParticleGrid;

        #endregion

        #region Nested Types: ParticleCell

        private struct ParticleCell
        {
            #region Fields

            public readonly ParticleIndex Particle;
            public readonly CellIndex CellIndex;

            #endregion

            #region Constructor

            public ParticleCell(in ParticleIndex particle, in CellIndex cellIndex)
            {
                Particle = particle;
                CellIndex = cellIndex;
            }

            #endregion
        }

        #endregion

        #region Nested Types: ParticleCellComparer

        private sealed class ParticleCellComparer : IComparer<ParticleCell>
        {
            /// <inheritdoc />
            public int Compare(ParticleCell x, ParticleCell y)
            {
                return x.CellIndex.CompareTo(y.CellIndex);
            }
        }

        #endregion

        #region Properties

        [NotNull, Browsable(true), TypeConverter(typeof (PluginTypeConverter<ICellIndexComputer>))]
        public ICellIndexComputer CellIndexComputer { get; set; } = new MortonCellIndex();

        #endregion

        #region Implementation of IParticleNeighborhood

        /// <inheritdoc />
        public bool HasNeighbors { get; private set; }

        /// <inheritdoc />
        public void SearchNeighbors(IParticleContext particleContext)
        {
            HasNeighbors = false;
            m_FluidParticleGrid = null;
            m_BoundaryParticleGrid = null;

            var boundingBox = ComputeBoundingBox(particleContext);

            var cellIndexComputer = CellIndexComputer;
            cellIndexComputer.Setup(in boundingBox, particleContext.Kernel.SupportLength);

            var performanceMeasurement = particleContext.PerformanceMeasurement;
            var indicesStopwatch = performanceMeasurement.Restart("Indices");
            var fluidParticleCellIndices = ComputeCellIndices(particleContext.Parallel, cellIndexComputer, particleContext.FluidParticles);
            indicesStopwatch.Stop();
            var sortAndReorderStopwatch = performanceMeasurement.Restart("Sorting and reordering particles");
            ParallelAlgorithms.Sort(fluidParticleCellIndices, new ParticleCellComparer());
            particleContext.ReorderFluidParticles(index => fluidParticleCellIndices[index].Particle, false);
            sortAndReorderStopwatch.Stop();
            var buildStopwatch = performanceMeasurement.Restart("Build grid");
            var fluidParticleGrid = m_FluidParticleGrid = BuildGrid(particleContext.Parallel, fluidParticleCellIndices);
            buildStopwatch.Stop();

            indicesStopwatch.Start();
            var boundaryParticleCellIndices = ComputeCellIndices(particleContext.Parallel, cellIndexComputer, particleContext.BoundaryParticles);
            indicesStopwatch.Stop();
            sortAndReorderStopwatch.Start();
            ParallelAlgorithms.Sort(boundaryParticleCellIndices, new ParticleCellComparer());
            particleContext.ReorderBoundaryParticles(index => boundaryParticleCellIndices[index].Particle, false);
            sortAndReorderStopwatch.Stop();
            buildStopwatch.Start();
            var boundaryParticleGrid = m_BoundaryParticleGrid = BuildGrid(particleContext.Parallel, boundaryParticleCellIndices);
            buildStopwatch.Stop();

            var queryGridStopwatch = performanceMeasurement.Restart("Query grid");
            SearchNeighbors(particleContext, cellIndexComputer, fluidParticleGrid, fluidParticleGrid, boundaryParticleGrid, particleContext.FluidParticles);
            SearchNeighbors(particleContext, cellIndexComputer, boundaryParticleGrid, fluidParticleGrid, boundaryParticleGrid, particleContext.BoundaryParticles);
            queryGridStopwatch.Stop();
            HasNeighbors = true;
        }

        /// <inheritdoc />
        public void SearchNeighbors(IParticleContext particleContext, Vector position, FluidParticleAction fluidParticleAction, BoundaryParticleAction boundaryParticleAction)
        {
            if (HasNeighbors)
            { 
                var kernel = particleContext.Kernel;
                var fluidParticles = particleContext.FluidParticles;
                var boundaryParticles = particleContext.BoundaryParticles;
                var fluidParticleGrid = m_FluidParticleGrid;
                var boundaryParticleGrid = m_BoundaryParticleGrid;
                var cellIndexComputer = CellIndexComputer;

                cellIndexComputer.ForEachNeighborCellIndex(cellIndexComputer.ComputeCellIndex(position), neighborCell =>
                {
                    if (fluidParticleGrid.TryGetValue(neighborCell, out var fluidParticleList))
                    {
                        foreach (var fIndex in fluidParticleList)
                        {
                            ref var f = ref fluidParticles[fIndex];
                            if (kernel.IsInRange(in position, f.Position))
                            {
                                fluidParticleAction(ref f);
                            }
                        }
                    }

                    if (boundaryParticleGrid.TryGetValue(neighborCell, out var boundaryParticleList))
                    {
                        foreach (var bIndex in boundaryParticleList)
                        {
                            ref var b = ref boundaryParticles[bIndex];
                            if (!kernel.IsInRange(in position, b.Position))
                            {
                                boundaryParticleAction(ref b);
                            }
                        }
                    }
                });
            }
        }

        #endregion

        #region Methods

        private static BoundingBox ComputeBoundingBox([NotNull] IParticleContext particleContext)
        {
            var fluidParticleBoundingBox
                = particleContext.ForEachFluidParticleReduceInParallel((ref FluidParticle f) => new BoundingBox(f.Position, f.Position),
                                                                       BoundingBox.Combine,
                                                                       BoundingBox.Invalid);

            var boundaryParticleBoundingBox
                = particleContext.ForEachBoundaryParticleReduceInParallel((ref BoundaryParticle b) => new BoundingBox(b.Position, b.Position),
                                                                          BoundingBox.Combine,
                                                                          BoundingBox.Invalid);

            return BoundingBox.Combine(fluidParticleBoundingBox, boundaryParticleBoundingBox);
        }

        [NotNull]
        private static ParticleCell[] ComputeCellIndices<T>([NotNull] IParallelContext parallel, [NotNull] ICellIndexComputer cellIndexComputer, [NotNull] T[] particles) where T : struct, IParticle
        {
            var cellIndices = new ParticleCell[particles.Length];
            parallel.ForEachRange(0, particles.Length, (startIndex, endIndex) =>
            {
                for (var i = startIndex; i < endIndex; ++i)
                {
                    cellIndices[i] = new ParticleCell(new ParticleIndex(i), cellIndexComputer.ComputeCellIndex(particles[i].Position));
                }
            });
            return cellIndices;
        }

        [NotNull]
        private static IEnumerable<Tuple<int, int>> CreateDistinctParticleCellPartitions([NotNull] IReadOnlyList<ParticleCell> particleCellIndices, int numberOfProcessors)
        {
            var list = new List<Tuple<int, int>>(numberOfProcessors);

            var chunkSize = (particleCellIndices.Count + numberOfProcessors - 1) / numberOfProcessors;
            if (chunkSize > 0)
            {
                var startIndex = 0;
                for (var p = 0; p < numberOfProcessors; ++p)
                {
                    var endIndex = Math.Min((p + 1) * chunkSize, particleCellIndices.Count) - 1;
                    var lastCellINdex = particleCellIndices[endIndex].CellIndex;
                    do
                    {
                        ++endIndex;
                    } while (endIndex < particleCellIndices.Count && lastCellINdex == particleCellIndices[endIndex].CellIndex);

                    list.Add(Tuple.Create(startIndex, endIndex));

                    if (endIndex >= particleCellIndices.Count)
                    {
                        break;
                    }

                    startIndex = endIndex;
                }
            }

            return list;
        }

        [NotNull]
        private static ConcurrentDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>> BuildGrid([NotNull] IParallelContext parallel, [NotNull] ParticleCell[] particleCellIndices)
        {
            var processors = parallel.NumThreads;
            var particleGrid = new ConcurrentDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>>(processors, particleCellIndices.Length);

            parallel.ForEach(CreateDistinctParticleCellPartitions(particleCellIndices, processors), 1, 1, range =>
            {
                var currentIndex = CellIndex.Invalid;
                List<ParticleIndex> currentList = null;

                for (var i = range.Item1; i < range.Item2; ++i)
                {
                    var particleCell = particleCellIndices[i];
                    if (currentIndex == particleCell.CellIndex)
                    {
                        currentList.Add(new ParticleIndex(i));
                    }
                    else
                    {
                        currentIndex = particleCell.CellIndex;
                        currentList = new List<ParticleIndex> {new ParticleIndex(i)};
                        particleGrid.TryAdd(currentIndex, currentList);
                    }
                }
            });

            return particleGrid;
        }

        private static void SearchNeighbors<T>(
            [NotNull] IParticleContext particleContext, [NotNull] ICellIndexComputer cellIndexComputer,
            [NotNull] IReadOnlyDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>> particleGrid, [NotNull] IReadOnlyDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>> fluidParticleGrid, [NotNull] IReadOnlyDictionary<CellIndex, IReadOnlyCollection<ParticleIndex>> boundaryParticleGrid,
            [NotNull] T[] particles) where T : struct, IParticle
        {
            var kernel = particleContext.Kernel;
            var fluidParticles = particleContext.FluidParticles;
            var boundaryParticles = particleContext.BoundaryParticles;

            particleContext.Parallel.ForEach(particleGrid, kvp =>
            {
                foreach (var iIndex in kvp.Value)
                {
                    NaiveParticleNeighborhood.ClearNeighborSets(ref particles[iIndex]);
                }

                cellIndexComputer.ForEachNeighborCellIndex(kvp.Key, neighborCell =>
                {
                    if (fluidParticleGrid.TryGetValue(neighborCell, out var fluidParticleList))
                    {
                        foreach (var iIndex in kvp.Value)
                        {
                            ref var i = ref particles[iIndex];

                            foreach (var fIndex in fluidParticleList)
                            {
                                ref var f = ref fluidParticles[fIndex];
                                if (!ParticleEquals.ReferenceEquals(ref i, ref f, iIndex, fIndex) && kernel.IsInRange(i.Position, f.Position))
                                {
                                    ((NeighborSet) i.FluidNeighbors).Add(fIndex);
                                }
                            }
                        }
                    }

                    if (boundaryParticleGrid.TryGetValue(neighborCell, out var boundaryParticleList))
                    {
                        foreach (var iIndex in kvp.Value)
                        {
                            ref var i = ref particles[iIndex];

                            foreach (var bIndex in boundaryParticleList)
                            {
                                ref var b = ref boundaryParticles[bIndex];
                                if (!ParticleEquals.ReferenceEquals(ref i, ref b, iIndex, bIndex) && kernel.IsInRange(i.Position, b.Position))
                                {
                                    ((NeighborSet) i.BoundaryNeighbors).Add(bIndex);
                                }
                            }
                        }
                    }
                });
            });
        }

        #endregion
    }
}
