﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.CompilerServices;
using System.Text;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.Kernel;
using TamiFlu.Core.Simulation.Neighborhood;
using TamiFlu.Core.Simulation.Solver;
using TamiFlu.Core.UI.Rendering.Color;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Neighborhood;
using TamiFlu.Interfaces.Rendering;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Kernel;
using TamiFlu.Interfaces.Simulation.Solver;

namespace TamiFlu.Core.Simulation
{
    [TypeConverter(typeof(ExpandableObjectConverter))]
    internal sealed class ParticleContext : IParticleContext, INotifyPropertyChanged
    {
        #region Fields

        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private IParallelContext m_Parallel = new ParallelContext();
        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private ITimeStep m_TimeStep = new TimeStep();
        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private IFluidSolver m_FluidSolver = new IISPHFluidSolver();
        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private IKernel m_Kernel = new CubicSplineKernel();
        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private IParticleNeighborhood m_Neighborhood = new UniformGridParticleNeighborhood();
        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private IParticleSerializer m_Serializer = new ParticleSerializer();
        [DebuggerBrowsable(DebuggerBrowsableState.Never)] private IParticleColorComputer m_ColorComputer = new VelocityColorComputer();

        #endregion

        #region Constructor

        public ParticleContext()
        {
            GlobalConstants.Instance.PropertyChanged += GlobalConstantChanged;
        }

        ~ParticleContext()
        {
            GlobalConstants.Instance.PropertyChanged -= GlobalConstantChanged;
        }

        #endregion

        #region Methods

        private void GlobalConstantChanged(object sender, [NotNull] PropertyChangedEventArgs e)
        {
            switch (e.PropertyName)
            {
                case "Spacing":
                    // spacing has changed, therefore the support length must be re-computed
                    Kernel.SupportFactor = Kernel.SupportFactor;
                    break;
            }
        }

        private void SetProperty<T>([CanBeNull] ref T field, [CanBeNull] T newValue, [CallerMemberName] string propertyName = null)
        {
            if (!EqualityComparer<T>.Default.Equals(field, newValue))
            {
                field = newValue;

                OnPropertyChanged(propertyName);

                // clear all performance measurements
                PerformanceMeasurement.Reset();
            }
        }

        #endregion

        #region Implementation of IParticleContext

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<IParallelContext>)), UsedImplicitly]
        public IParallelContext Parallel
        {
            get => m_Parallel;
            set => SetProperty(ref m_Parallel, value);
        }

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<ITimeStep>)), UsedImplicitly]
        public ITimeStep TimeStep
        {
            get => m_TimeStep;
            set => SetProperty(ref m_TimeStep, value);
        }

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<IFluidSolver>)), UsedImplicitly]
        public IFluidSolver FluidSolver
        {
            get => m_FluidSolver;
            set => SetProperty(ref m_FluidSolver, value);
        }

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<IKernel>)), UsedImplicitly]
        public IKernel Kernel
        {
            get => m_Kernel;
            set => SetProperty(ref m_Kernel, value);
        }

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<IParticleNeighborhood>)), UsedImplicitly]
        public IParticleNeighborhood Neighborhood
        {
            get => m_Neighborhood;
            set => SetProperty(ref m_Neighborhood, value);
        }

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<IParticleSerializer>)), UsedImplicitly]
        public IParticleSerializer Serializer
        {
            get => m_Serializer;
            set => SetProperty(ref m_Serializer, value);
        }

        /// <inheritdoc />
        [Browsable(true), TypeConverter(typeof(PluginTypeConverter<IParticleColorComputer>)), UsedImplicitly]
        public IParticleColorComputer ColorComputer
        {
            get => m_ColorComputer;
            set => SetProperty(ref m_ColorComputer, value);
        }

        /// <inheritdoc />
        [Browsable(false)]
        public IPerformanceMeasurement PerformanceMeasurement { get; } = new PerformanceMeasurement();

        /// <inheritdoc />
        [Browsable(false)]
        public FluidParticle[] FluidParticles { get; private set; } = Array.Empty<FluidParticle>();

        /// <inheritdoc />
        [Browsable(false)]
        public BoundaryParticle[] BoundaryParticles { get; private set; } = Array.Empty<BoundaryParticle>();

        /// <inheritdoc />
        public ParticleIndex? PickedFluidParticle { get; set; }

        /// <inheritdoc />
        public ParticleIndex? PickedBoundaryParticle { get; set; }

        /// <inheritdoc />
        public void Clear()
        {
            FluidParticles = Array.Empty<FluidParticle>();
            BoundaryParticles = Array.Empty<BoundaryParticle>();

            TimeStep.ResetCurrentTime();
        }

        public ParticleIndex CreateFluidParticles(FluidParticlePhysicsProperties properties, int count)
        {
            if (properties is null)
            {
                throw new ArgumentNullException(nameof(properties));
            }

            var oldFluidParticles = FluidParticles;
            var newFluidParticles = new FluidParticle[oldFluidParticles.Length + count];

            Parallel.ForEachRange(0, oldFluidParticles.Length, (startIndex, endIndex) =>
            {
                Array.Copy(oldFluidParticles, startIndex, newFluidParticles, startIndex, endIndex - startIndex);
            });

            Parallel.ForEachRange(oldFluidParticles.Length, newFluidParticles.Length, (startIndex, endIndex) =>
            {
                for (var f = startIndex; f < endIndex; ++f)
                {
                    newFluidParticles[f] = new FluidParticle(properties);
                }
            });

            FluidParticles = newFluidParticles;
            return new ParticleIndex(oldFluidParticles.Length);
        }

        /// <inheritdoc />
        public ParticleIndex CreateBoundaryParticles(BoundaryParticlePhysicsProperties properties, int count)
        {
            if (properties is null)
            {
                throw new ArgumentNullException(nameof(properties));
            }

            var oldBoundaryParticles = BoundaryParticles;
            var newBoundaryParticles = new BoundaryParticle[oldBoundaryParticles.Length + count];

            Parallel.ForEachRange(0, oldBoundaryParticles.Length, (startIndex, endIndex) =>
            {
                Array.Copy(oldBoundaryParticles, startIndex, newBoundaryParticles, startIndex, endIndex - startIndex);
            });

            Parallel.ForEachRange(oldBoundaryParticles.Length, newBoundaryParticles.Length, (startIndex, endIndex) =>
            {
                for (var b = startIndex; b < endIndex; ++b)
                {
                    newBoundaryParticles[b] = new BoundaryParticle(properties);
                }
            });

            BoundaryParticles = newBoundaryParticles;
            return new ParticleIndex(oldBoundaryParticles.Length);
        }

        /// <inheritdoc />
        public void ReorderFluidParticles(Func<ParticleIndex, ParticleIndex> action, bool ignoreAttributes)
        {
            var pickedFluidParticle = PickedFluidParticle;
            FluidParticles = ReorderParticles(FluidParticles, action, ref pickedFluidParticle, ignoreAttributes, (ref FluidParticle f) => f.Properties, (ref FluidParticle f, FluidParticlePhysicsProperties properties) => f.Properties = properties);
            PickedFluidParticle = pickedFluidParticle;
        }

        /// <inheritdoc />
        public void ReorderBoundaryParticles(Func<ParticleIndex, ParticleIndex> action, bool ignoreAttributes)
        {
            var pickedBoundaryParticle = PickedBoundaryParticle;
            BoundaryParticles = ReorderParticles(BoundaryParticles, action, ref pickedBoundaryParticle, ignoreAttributes, (ref BoundaryParticle b) => b.Properties, (ref BoundaryParticle b, BoundaryParticlePhysicsProperties properties) => b.Properties = properties);
            PickedBoundaryParticle = pickedBoundaryParticle;
        }

        private delegate TProperties PropertiesGetter<T, out TProperties>(ref T particle) where T : struct, IParticle;
        private delegate void PropertiesSetter<T, in TProperties>(ref T particle, TProperties property) where T : struct, IParticle;

        [NotNull]
        private T[] ReorderParticles<T, TProperties>([NotNull] T[] particles, [NotNull] Func<ParticleIndex, ParticleIndex> action, ref ParticleIndex? pickedParticle, bool ignoreAttributes, [NotNull] PropertiesGetter<T, TProperties> getProperties, [NotNull] PropertiesSetter<T, TProperties> setProperties) where T : struct, IParticle
        {
            if (particles is null)
            {
                throw new ArgumentNullException(nameof(particles));
            }

            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            var searchPickedParticle = pickedParticle.HasValue;
            var localPickedParticle = pickedParticle.GetValueOrDefault();

            T[] newParticles;
            if (ignoreAttributes)
            {
                newParticles = particles;
                var positions = new Vector[particles.Length];
                var velocities = new Vector[particles.Length];
                var properties = new TProperties[particles.Length];
                Parallel.ForEachRange(0, particles.Length, (startIndex, endIndex) =>
                {
                    for (var i = startIndex; i < endIndex; ++i)
                    {
                        var newIndex = new ParticleIndex(i);
                        var oldIndex = action(newIndex);

                        // copy only important attributes to new array
                        ref var particle = ref particles[oldIndex];
                        positions[i] = particle.Position;
                        velocities[i] = particle.Velocity;
                        properties[i] = getProperties(ref particle);

                        // assign correct index to picked particle
                        // NOTE this check is thread-safe since its condition
                        // is only valid for one particle
                        if (searchPickedParticle && localPickedParticle == oldIndex)
                        {
                            searchPickedParticle = false;
                            localPickedParticle = newIndex;
                        }
                    }
                });

                Parallel.ForEachRange(0, particles.Length, (startIndex, endIndex) =>
                {
                    for (var i = startIndex; i < endIndex; ++i)
                    {
                        ref var particle = ref particles[i];
                        particle.Position = positions[i];
                        particle.Velocity = velocities[i];
                        setProperties(ref particle, properties[i]);
                    }
                });
            }
            else
            {
                newParticles = new T[particles.Length];
                Parallel.ForEachRange(0, particles.Length, (startIndex, endIndex) =>
                {
                    for (var i = startIndex; i < endIndex; ++i)
                    {
                        var newIndex = new ParticleIndex(i);
                        var oldIndex = action(newIndex);

                        // copy particle to new array
                        newParticles[i] = particles[oldIndex];

                        // assign correct index to picked particle
                        // NOTE this check is thread-safe since its condition
                        // is only valid for one particle
                        if (searchPickedParticle && localPickedParticle == oldIndex)
                        {
                            searchPickedParticle = false;
                            localPickedParticle = newIndex;
                        }
                    }
                });
            }

            if (pickedParticle.HasValue)
            {
                pickedParticle = localPickedParticle;
            }

            return newParticles;
        }

        /// <inheritdoc />
        public void ForEachFluidParticle(FluidParticleAction action)
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            var fluidParticles = FluidParticles;
            for (var f = ParticleIndex.Zero; f < fluidParticles.Length; ++f)
            {
                action(ref fluidParticles[f]);
            }
        }

        /// <inheritdoc />
        public void ForEachFluidParticle(FluidParticleIndexAction action)
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            var fluidParticles = FluidParticles;
            for (var f = ParticleIndex.Zero; f < fluidParticles.Length; ++f)
            {
                action(ref fluidParticles[f], f);
            }
        }

        /// <inheritdoc />
        public void ForEachBoundaryParticle(BoundaryParticleAction action)
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            var boundaryParticles = BoundaryParticles;
            for (var b = ParticleIndex.Zero; b < boundaryParticles.Length; ++b)
            {
                action(ref boundaryParticles[b]);
            }
        }

        /// <inheritdoc />
        public void ForEachBoundaryParticle(BoundaryParticleIndexAction action)
        {
            if (action is null)
            {
                throw new ArgumentNullException(nameof(action));
            }

            var boundaryParticles = BoundaryParticles;
            for (var b = ParticleIndex.Zero; b < boundaryParticles.Length; ++b)
            {
                action(ref boundaryParticles[b], b);
            }
        }

        /// <inheritdoc />
        public void ForEachFluidParticleInParallel(FluidParticleAction action)
        {
            var fluidParticles = FluidParticles;
            Parallel.ForEachRange(0, fluidParticles.Length, (startIndex, endIndex) =>
            {
                for (var f = startIndex; f < endIndex; ++f)
                {
                    action(ref fluidParticles[f]);
                }
            });
        }

        /// <inheritdoc />
        public void ForEachFluidParticleInParallel(FluidParticleIndexAction action)
        {
            var fluidParticles = FluidParticles;
            Parallel.ForEachRange(0, fluidParticles.Length, (startIndex, endIndex) =>
            {
                for (var f = startIndex; f < endIndex; ++f)
                {
                    action(ref fluidParticles[f], new ParticleIndex(f));
                }
            });
        }

        /// <inheritdoc />
        public void ForEachBoundaryParticleInParallel(BoundaryParticleAction action)
        {
            var boundaryParticles = BoundaryParticles;
            Parallel.ForEachRange(0, boundaryParticles.Length, (startIndex, endIndex) =>
            {
                for (var b = startIndex; b < endIndex; ++b)
                {
                    action(ref boundaryParticles[b]);
                }
            });
        }

        /// <inheritdoc />
        public void ForEachBoundaryParticleInParallel(BoundaryParticleIndexAction action)
        {
            var boundaryParticles = BoundaryParticles;
            Parallel.ForEachRange(0, boundaryParticles.Length, (startIndex, endIndex) =>
            {
                for (var b = startIndex; b < endIndex; ++b)
                {
                    action(ref boundaryParticles[b], new ParticleIndex(b));
                }
            });
        }

        /// <inheritdoc />
        public T ForEachFluidParticleReduceInParallel<T>(FluidParticleFunction<T> action, Func<T, T, T> reduction, T initValue) where T : struct
        {
            var fluidParticles = FluidParticles;
            return Parallel.ReduceRange(0, fluidParticles.Length, f => action(ref fluidParticles[f]), reduction, initValue);
        }

        /// <inheritdoc />
        public T ForEachFluidParticleReduceInParallel<T>(FluidParticleIndexFunction<T> action, Func<T, T, T> reduction, T initValue) where T : struct
        {
            var fluidParticles = FluidParticles;
            return Parallel.ReduceRange(0, fluidParticles.Length, f => action(ref fluidParticles[f], new ParticleIndex(f)), reduction, initValue);
        }

        /// <inheritdoc />
        public T ForEachBoundaryParticleReduceInParallel<T>(BoundaryParticleFunction<T> action, Func<T, T, T> reduction, T initValue) where T : struct
        {
            var boundaryParticles = BoundaryParticles;
            return Parallel.ReduceRange(0, boundaryParticles.Length, b => action(ref boundaryParticles[b]), reduction, initValue);
        }

        /// <inheritdoc />
        public T ForEachBoundaryParticleReduceInParallel<T>(BoundaryParticleIndexFunction<T> action, Func<T, T, T> reduction, T initValue) where T : struct
        {
            var boundaryParticles = BoundaryParticles;
            return Parallel.ReduceRange(0, boundaryParticles.Length, b => action(ref boundaryParticles[b], new ParticleIndex(b)), reduction, initValue);
        }
        
        /// <inheritdoc />
        public string GetDebugText()
        {
            var stringBuilder = new StringBuilder();
            stringBuilder.AppendLine($"Fluid particles: {FluidParticles.Length}");
            stringBuilder.AppendLine($"Boundary particles: {BoundaryParticles.Length}");

            foreach (var obj in new object[]
                                {
                                    Parallel,
                                    TimeStep,
                                    FluidSolver,
                                    Kernel,
                                    Neighborhood,
                                    Serializer,
                                    ColorComputer
                                })
            {
                if (obj is IDebugDrawable drawable)
                {
                    drawable.AppendText(stringBuilder, this);
                }
            }

            if (PickedFluidParticle.HasValue)
            {
                ref readonly var f = ref FluidParticles[PickedFluidParticle.Value];
                stringBuilder.AppendLine();
                stringBuilder.AppendLine("Picked fluid particle");
                stringBuilder.AppendLine($"Index: {PickedFluidParticle.Value}");
                stringBuilder.AppendLine($"Position: {f.Position}");
                stringBuilder.AppendLine($"Velocity: {f.Velocity}\u2009m/s");
                stringBuilder.AppendLine($"Acceleration: {f.Acceleration}\u2009m/s²");
                stringBuilder.AppendLine($"Pressure: {f.Pressure}\u2009Pa");
                stringBuilder.AppendLine($"Density: {f.Density}\u2009kg/m³");
                stringBuilder.AppendLine($"Volume: {f.Volume}\u2009m³");
                stringBuilder.AppendLine($"Fluid neighbors: {f.FluidNeighbors?.Count ?? 0}");
                stringBuilder.AppendLine($"Boundary neighbors: {f.BoundaryNeighbors?.Count ?? 0}");
                if (!(f.UserObject is null))
                {
                    stringBuilder.AppendLine(f.UserObject.ToString());
                }
            }
            else if (PickedBoundaryParticle.HasValue)
            {
                ref readonly var b = ref BoundaryParticles[PickedBoundaryParticle.Value];
                stringBuilder.AppendLine();
                stringBuilder.AppendLine("Picked boundary particle");
                stringBuilder.AppendLine($"Index: {PickedBoundaryParticle.Value}");
                stringBuilder.AppendLine($"Position: {b.Position}");
                stringBuilder.AppendLine($"Velocity: {b.Velocity}\u2009m/s");
                stringBuilder.AppendLine($"Volume: {b.Volume}\u2009m³");
                stringBuilder.AppendLine($"Fluid neighbors: {b.FluidNeighbors?.Count ?? 0}");
                stringBuilder.AppendLine($"Boundary neighbors: {b.BoundaryNeighbors?.Count ?? 0}");
                if (!(b.UserObject is null))
                {
                    stringBuilder.AppendLine(b.UserObject.ToString());
                }
            }

            if (PerformanceMeasurement.Count != 0)
            {
                stringBuilder.AppendLine();
                stringBuilder.AppendLine(((PerformanceMeasurement) PerformanceMeasurement).PrettyFormat());
            }

            return stringBuilder.ToString();
        }

        #endregion

        #region Implementation of INotifyPropertyChanged

        /// <inheritdoc />
        public event PropertyChangedEventHandler PropertyChanged;

        [NotifyPropertyChangedInvocator]
        private void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion
    }
}
