﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using JetBrains.Annotations;
using System;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.Simulation
{
    public sealed class ParticleSource : IParticleSource
    {
        #region Methods

        private static int TotalCount([NotNull] int[] numParticlesToGeneratePerDimension)
        {
            var totalCount = 1;
            foreach (var numParticlesToGenerate in numParticlesToGeneratePerDimension)
            {
                totalCount *= numParticlesToGenerate;
            }
            return totalCount;
        }

        [NotNull]
        private static int[] ComputeNumberOfParticles(Vector size)
        {
            var numParticlesToGenerate = size / GlobalConstants.Instance.Spacing;
            var numParticlesToGeneratePerDimension = new int[3];
            for (var i = 0; i < Vector.Dimensions; ++i)
            {
                numParticlesToGeneratePerDimension[i] = (int)Math.Floor(numParticlesToGenerate[i]);
            }
            for (var i = Vector.Dimensions; i < numParticlesToGeneratePerDimension.Length; ++i)
            {
                numParticlesToGeneratePerDimension[i] = 1;
            }

            return numParticlesToGeneratePerDimension;
        }

        private static void SampleParticles<T>([NotNull] T[] particles, ParticleIndex startIndex, [NotNull] int[] numParticlesToGeneratePerDimension, Vector center, Vector size, [NotNull] ParticleAction<T> action) where T : struct, IParticle
        {
            var startPosition = center - size / 2.0f;

            for (var y = 0; y < numParticlesToGeneratePerDimension[1]; ++y)
            {
                for (var x = 0; x < numParticlesToGeneratePerDimension[0]; ++x)
                {
                    ref var particle = ref particles[startIndex++];
                    particle.Position = startPosition + new Vector(x, y) * GlobalConstants.Instance.Spacing;
                    action(ref particle);
                }
            }
        }

        #endregion

        #region Implementation of IParticleSource

        /// <inheritdoc />
        public void SampleFluidParticleBlock(IParticleContext particleContext, Vector center, Vector size, FluidParticlePhysicsProperties properties)
        {
            var numParticlesToGeneratePerDimension = ComputeNumberOfParticles(size);
            var startIndex = particleContext.CreateFluidParticles(properties, TotalCount(numParticlesToGeneratePerDimension));
            SampleParticles(particleContext.FluidParticles, startIndex, numParticlesToGeneratePerDimension, center, size, (ref FluidParticle f) =>
            {
            });
        }

        /// <inheritdoc />
        public void SampleBoundaryParticleBlock(IParticleContext particleContext, Vector center, Vector size, BoundaryParticlePhysicsProperties properties)
        {
            var numParticlesToGeneratePerDimension = ComputeNumberOfParticles(size);
            var startIndex = particleContext.CreateBoundaryParticles(properties, TotalCount(numParticlesToGeneratePerDimension));
            SampleParticles(particleContext.BoundaryParticles, startIndex, numParticlesToGeneratePerDimension, center, size, (ref BoundaryParticle b) =>
            {
            });
        }

        #endregion
    }
}
