﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.Simulation
{
    internal sealed class PerformanceMeasurement : IPerformanceMeasurement
    {
        #region Nested Types: StopwatchNode

        private sealed class StopwatchNode
        {
            [NotNull] public readonly Stopwatch Value;
            [CanBeNull] public StopwatchNode Parent;

            public StopwatchNode()
            {
                Value = new Stopwatch();
            }
        }

        #endregion

        #region Fields

        [NotNull] private readonly object m_SyncObject = new object();
        [NotNull] private readonly Dictionary<string, StopwatchNode> m_Stopwatches = new Dictionary<string, StopwatchNode>();
        [NotNull, ItemNotNull] private readonly Stack<StopwatchNode> m_Stack = new Stack<StopwatchNode>();
        [CanBeNull] private StopwatchNode m_Current;

        #endregion

        #region Methods

        [NotNull]
        private Stopwatch Create([NotNull] string name)
        {
            lock (m_SyncObject)
            {
                if (!m_Stopwatches.TryGetValue(name, out var value))
                {
                    m_Stopwatches.Add(name, value = new StopwatchNode());
                }

                value.Parent = m_Stack.Count != 0 ? m_Stack.Peek() : null;
                m_Current = value;
                return value.Value;
            }
        }

        private static void PrettyFormat([NotNull] StringBuilder stringBuilder, TimeSpan timeSpan)
        {
            const double millisPerSecond = 1000.0;
            const double millisPerMinute = millisPerSecond * 60.0;

            var totalMillis = timeSpan.TotalMilliseconds;
            if (totalMillis > millisPerMinute)
            {
                var minutes = (int) Math.Truncate(totalMillis / millisPerMinute);
                stringBuilder.Append($" {minutes}\u2009min");
                totalMillis -= minutes * millisPerMinute;
            }

            if (totalMillis > 10.0 * millisPerSecond)
            {
                var seconds = (int) Math.Truncate(totalMillis / millisPerSecond);
                stringBuilder.Append($" {seconds}\u2009s");
                totalMillis -= seconds * millisPerMinute;
            }

            stringBuilder.Append($" {totalMillis}\u2009ms");
        }

        [NotNull]
        public string PrettyFormat()
        {
            var stringBuilder = new StringBuilder();

            foreach (var kvp in m_Stopwatches)
            {
                var parentCount = 0;
                var parent = kvp.Value.Parent;
                while (parent != null)
                {
                    ++parentCount;
                    parent = parent.Parent;
                }

                stringBuilder.Append(' ', 2*parentCount);
                stringBuilder.Append(kvp.Key);
                stringBuilder.Append(':');
                PrettyFormat(stringBuilder, kvp.Value.Value.Elapsed);
                stringBuilder.AppendLine();
            }

            return stringBuilder.ToString();
        }

        #endregion

        #region Implementation of IPerformanceMeasurement
        
        /// <inheritdoc />
        public int Count => m_Stopwatches.Count;

        /// <inheritdoc />
        public IEnumerable<string> Names => m_Stopwatches.Keys;

        /// <inheritdoc />
        public IEnumerable<Stopwatch> Stopwatches => m_Stopwatches.Values.Select(x => x.Value);

        /// <inheritdoc />
        public Stopwatch Current => m_Current?.Value;

        /// <inheritdoc />
        public TimeSpan Elapsed(string name)
        {
            lock (m_SyncObject)
            {
                if (m_Stopwatches.TryGetValue(name, out var value))
                {
                    return value.Value.Elapsed;
                }
            }

            return TimeSpan.Zero;
        }

        /// <inheritdoc />
        public void Push()
        {
            lock (m_SyncObject)
            {
                var current = m_Current;
                if (!(current is null))
                {
                    m_Stack.Push(current);
                }
            }
        }

        /// <inheritdoc />
        public void Pop()
        {
            lock (m_SyncObject)
            {
                if (m_Stack.Count != 0)
                {
                    m_Stack.Pop();
                }
            }
        }

        /// <inheritdoc />
        public void Reset()
        {
            lock (m_SyncObject)
            {
                m_Stopwatches.Clear();
                m_Stack.Clear();
                m_Current = null;
            }
        }

        /// <inheritdoc />
        public Stopwatch Reset(string name)
        {
            var value = Create(name);
            value.Reset();
            return value;
        }

        /// <inheritdoc />
        public Stopwatch Restart(string name)
        {
            var value = Create(name);
            value.Restart();
            return value;
        }

        /// <inheritdoc />
        public Stopwatch Start(string name)
        {
            var value = Create(name);
            value.Start();
            return value;
        }

        /// <inheritdoc />
        public void Measure(string name, Action action)
        {
            var stopwatch = Restart(name);
            Push();
            action();
            Pop();
            stopwatch.Stop();
        }

        /// <inheritdoc />
        public T Measure<T>(string name, Func<T> action)
        {
            var stopwatch = Restart(name);
            Push();
            var result = action();
            Pop();
            stopwatch.Stop();
            return result;
        }

        #endregion

        #region Implementation of IEnumerable

        /// <inheritdoc />
        public IEnumerator<KeyValuePair<string, Stopwatch>> GetEnumerator()
        {
            return m_Stopwatches.Select(x => new KeyValuePair<string, Stopwatch>(x.Key, x.Value.Value)).GetEnumerator();
        }

        /// <inheritdoc />
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion
    }
}
