﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion

using System;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Solver;

namespace TamiFlu.Core.Simulation.Solver.Boundary
{
    internal abstract class ABoundaryHandling : IBoundaryHandling
    {
        #region Methods

        protected static void ComputeBoundaryVolume([NotNull] IParticleContext particleContext)
        {
            var kernel = particleContext.Kernel;
            var selfKernelValue = kernel.ComputeValue(Vector.Zero, Vector.Zero);
            var boundaryParticles = particleContext.BoundaryParticles;

            // compute the volume of boundary particles
            particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b) =>
            {
                var numberDensity = selfKernelValue;
                foreach (var bbIndex in b.BoundaryNeighbors)
                {
                    ref readonly var bb = ref boundaryParticles[bbIndex];
                    numberDensity += kernel.ComputeValue(b.Position, bb.Position);
                }

                b.Volume = 1f / numberDensity;
            });
        }

        protected static void ComputeGradientAntisymmetric([NotNull] IParticleContext particleContext, [NotNull] GetFluidScalar getFluidScalar, [NotNull] GetBoundaryScalar getBoundaryScalar, [CanBeNull] FluidGradientEndAction fluidAction, [CanBeNull] BoundaryGradientEndAction boundaryAction)
        {
            var kernel = particleContext.Kernel;
            var fluidParticles = particleContext.FluidParticles;
            var boundaryParticles = particleContext.BoundaryParticles;

            if (!(fluidAction is null))
            {
                particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
                {
                    var fValue = getFluidScalar(in f, in fIndex);
                    var gradient = Vector.Zero;

                    foreach (var ffIndex in f.FluidNeighbors)
                    {
                        ref readonly var ff = ref fluidParticles[ffIndex];
                        gradient += (ff.Volume * (fValue + getFluidScalar(in ff, in ffIndex))) * kernel.ComputeGradient(f.Position, ff.Position);
                    }

                    foreach (var fbIndex in f.BoundaryNeighbors)
                    {
                        ref readonly var fb = ref boundaryParticles[fbIndex];
                        gradient += (fb.Volume * (fValue + getBoundaryScalar(in fb, in fbIndex))) * kernel.ComputeGradient(f.Position, fb.Position);
                    }

                    fluidAction(ref f, fIndex, in gradient);
                });
            }

            if (!(boundaryAction is null))
            {
                particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b, in ParticleIndex bIndex) =>
                {
                    var bValue = getBoundaryScalar(in b, in bIndex);
                    var gradient = Vector.Zero;

                    foreach (var bfIndex in b.FluidNeighbors)
                    {
                        ref readonly var bf = ref fluidParticles[bfIndex];
                        gradient += (bf.Volume * (bValue + getFluidScalar(in bf, in bfIndex))) * kernel.ComputeGradient(b.Position, bf.Position);
                    }

                    foreach (var bbIndex in b.BoundaryNeighbors)
                    {
                        ref readonly var bb = ref boundaryParticles[bbIndex];
                        gradient += (bb.Volume * (bValue + getBoundaryScalar(in bb, in bbIndex))) * kernel.ComputeGradient(b.Position, bb.Position);
                    }

                    boundaryAction(ref b, bIndex, in gradient);
                });
            }
        }

        protected static void ComputeDivergenceSymmetric([NotNull] IParticleContext particleContext, [NotNull] GetFluidVector getFluidVector, [NotNull] GetBoundaryVector getBoundaryVector, [CanBeNull] FluidDivergenceEndAction fluidAction, [CanBeNull] BoundaryDivergenceEndAction boundaryAction)
        {
            var kernel = particleContext.Kernel;
            var fluidParticles = particleContext.FluidParticles;
            var boundaryParticles = particleContext.BoundaryParticles;

            if (!(fluidAction is null))
            {
                particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
                {
                    var fValue = getFluidVector(in f, in fIndex);
                    var divergence = 0f;

                    foreach (var ffIndex in f.FluidNeighbors)
                    {
                        ref readonly var ff = ref fluidParticles[ffIndex];
                        divergence += ff.Volume * Vector.Dot(getFluidVector(in ff, in ffIndex) - fValue, kernel.ComputeGradient(f.Position, ff.Position));
                    }

                    foreach (var fbIndex in f.BoundaryNeighbors)
                    {
                        ref readonly var fb = ref boundaryParticles[fbIndex];
                        divergence += fb.Volume * Vector.Dot(getBoundaryVector(in fb, in fbIndex) - fValue, kernel.ComputeGradient(f.Position, fb.Position));
                    }

                    fluidAction(ref f, fIndex, divergence);
                });
            }

            if (!(boundaryAction is null))
            {
                particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b, in ParticleIndex bIndex) =>
                {
                    var bValue = getBoundaryVector(in b, in bIndex);
                    var divergence = 0f;

                    foreach (var bfIndex in b.FluidNeighbors)
                    {
                        ref readonly var bf = ref fluidParticles[bfIndex];
                        divergence += bf.Volume * Vector.Dot(getFluidVector(in bf, in bfIndex) - bValue, kernel.ComputeGradient(b.Position, bf.Position));
                    }

                    foreach (var bbIndex in b.BoundaryNeighbors)
                    {
                        ref readonly var bb = ref boundaryParticles[bbIndex];
                        divergence += bb.Volume * Vector.Dot(getBoundaryVector(in bb, in bbIndex) - bValue, kernel.ComputeGradient(b.Position, bb.Position));
                    }

                    boundaryAction(ref b, bIndex, divergence);
                });
            }
        }

        [NotNull]
        internal static T[] EnsureLength<T>([CanBeNull] ref T[] array, int length)
        {
            if (array is null || array.Length != length)
            {
                array = length == 0 ? Array.Empty<T>() : new T[length];
            }

            return array;
        }

        #endregion

        #region Implementation of IBoundaryHandling

        /// <inheritdoc />
        public void ComputeVolume(IParticleContext particleContext)
        {
            var kernel = particleContext.Kernel;
            var selfKernelValue = kernel.ComputeValue(Vector.Zero, Vector.Zero);
            var restVolume = GlobalConstants.Instance.RestVolume;
            var fluidParticles = particleContext.FluidParticles;
            var boundaryParticles = particleContext.BoundaryParticles;

            // compute the volume of boundary particles
            ComputeBoundaryVolume(particleContext);

            // compute the density of fluid particles
            particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) =>
            {
                var numberDensity = selfKernelValue;

                foreach (var ffIndex in f.FluidNeighbors)
                {
                    ref readonly var ff = ref fluidParticles[ffIndex];
                    numberDensity += kernel.ComputeValue(f.Position, ff.Position);
                }

                foreach (var fbIndex in f.BoundaryNeighbors)
                {
                    ref readonly var fb = ref boundaryParticles[fbIndex];
                    numberDensity += (fb.Volume / restVolume) * kernel.ComputeValue(f.Position, fb.Position);
                }

                f.Volume = 1f / numberDensity;
            });
        }

        /// <inheritdoc />
        public void ComputeGradient(IParticleContext particleContext, GetFluidScalar getFluidScalar, GetBoundaryScalar getBoundaryScalar, FluidGradientEndAction fluidAction, BoundaryGradientEndAction boundaryAction)
        {
            ComputeGradientAntisymmetric(particleContext, getFluidScalar, getBoundaryScalar, fluidAction, boundaryAction);
        }

        /// <inheritdoc />
        public void ComputeDivergence(IParticleContext particleContext, GetFluidVector getFluidVector, GetBoundaryVector getBoundaryVector, FluidDivergenceEndAction fluidAction, BoundaryDivergenceEndAction boundaryAction)
        {
            ComputeDivergenceSymmetric(particleContext, getFluidVector, getBoundaryVector, fluidAction, boundaryAction);
        }
        
        #endregion
    }
}
