﻿using System.ComponentModel;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.Forces;
using TamiFlu.Core.Simulation.SystemSolver;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.Simulation.Solver.Boundary
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("Pressure Extrapolation with SPH [Adami et al. 2012]", "1.0", "Stefan Band"), UsedImplicitly]
    internal sealed class AdamiBoundaryHandling : ABoundaryHandling, IIterativeBoundaryHandling
    {
        #region Implementation of IIterativeBoundaryHandling

        /// <inheritdoc />
        public ComputeDiagonalElements CreateComputeDiagonalElementFunc(IParticleContext particleContext)
        {
            return CreateStaticComputeDiagonalElementFunc(particleContext);
        }
        
        /// <inheritdoc />
        public ComputeSourceTerms CreateComputeSourceTermFunc(IParticleContext particleContext)
        {
            return AkinciBoundaryHandling.CreateStaticComputeSourceTermFunc(this, particleContext);
        }

        /// <inheritdoc />
        public InitializeVector CreateInitializeVectorFunc(IParticleContext particleContext, float factor)
        {
            return AkinciBoundaryHandling.CreateStaticInitializeVectorFunc(this, particleContext, factor);
        }

        /// <inheritdoc />
        public ComputeMatrixVectorProduct CreateComputeMatrixVectorProductFunc(IParticleContext particleContext, ref Vector[] pressureAccelerations)
        {
            var fluidPressureAccelerations = EnsureLength(ref pressureAccelerations, particleContext.FluidParticles.Length);

            void ComputeMatrixVectorProduct(float[] vectors, ref float[] matrixVectorProducts)
            {
                var values = EnsureLength(ref matrixVectorProducts, particleContext.FluidParticles.Length);
                var timeStep = particleContext.TimeStep.TimeStepValue;

                // compute boundary pressures
                var boundaryPressures = new float[particleContext.BoundaryParticles.Length];
                {
                    var fluidParticles = particleContext.FluidParticles;
                    var gravity = GetGravityVector(particleContext);
                    var kernel = particleContext.Kernel;
                    particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b, in ParticleIndex bIndex) =>
                    {
                        var sumKernel = 0f;
                        var sumPressure = 0f;

                        foreach (var bfIndex in b.FluidNeighbors)
                        {
                            ref readonly var bf = ref fluidParticles[bfIndex];
                            var kernelValue = bf.Volume * kernel.ComputeValue(b.Position, bf.Position);
                            sumPressure += bf.Pressure * kernelValue;
                            sumPressure += bf.Density * Vector.Dot(gravity - b.Velocity/timeStep, b.Position - bf.Position) * kernelValue;
                            sumKernel += kernelValue;
                        }

                        boundaryPressures[bIndex] = sumKernel > 0f ? (sumPressure / sumKernel) : 0f;
                    });
                }

                // compute pressure accelerations
                ComputeGradient(particleContext,
                                (in FluidParticle f, in ParticleIndex fIndex) => vectors[fIndex],
                                (in BoundaryParticle b, in ParticleIndex bIndex) => boundaryPressures[bIndex],
                                (ref FluidParticle f, in ParticleIndex fIndex, in Vector pressureGradient) => fluidPressureAccelerations[fIndex] = pressureGradient / -f.Density,
                                null);

                // compute divergence of pressure acceleration
                ComputeDivergence(particleContext,
                                  (in FluidParticle f, in ParticleIndex fIndex) => fluidPressureAccelerations[fIndex],
                                  (in BoundaryParticle b, in ParticleIndex bIndex) => Vector.Zero,
                                  (ref FluidParticle f, in ParticleIndex fIndex, float pressureAccelerationDivergence) => values[fIndex] = (timeStep * timeStep) * pressureAccelerationDivergence,
                                  null);
            }

            return ComputeMatrixVectorProduct;
        }

        #endregion

        #region Methods

        private static Vector GetGravityVector([NotNull] IParticleContext particleContext)
        {
            return ((GravityForce) ((IISPHFluidSolver) particleContext.FluidSolver).GravityForce).Gravity;
        }

        [NotNull]
        internal static ComputeDiagonalElements CreateStaticComputeDiagonalElementFunc(IParticleContext particleContext)
        {
            void ComputeDiagonalElement(ref float[] diagonalElements)
            {
                var fluidParticles = particleContext.FluidParticles;
                var boundaryParticles = particleContext.BoundaryParticles;
                var values = EnsureLength(ref diagonalElements, fluidParticles.Length);
                var timeStep = particleContext.TimeStep.TimeStepValue;
                var kernel = particleContext.Kernel;

                // compute diagonal elements
                particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
                {
                   var sumGradient = Vector.Zero;
                   var sumGradientSquared = 0f;

                   foreach (var ffIndex in f.FluidNeighbors)
                   {
                       ref readonly var ff = ref fluidParticles[ffIndex];
                       var gradient = kernel.ComputeGradient(f.Position, ff.Position);
                       sumGradient += ff.Volume * gradient;
                       sumGradientSquared += ff.Volume / ff.Density * Vector.LengthSquared(in gradient);
                   }

                   foreach (var fbIndex in f.BoundaryNeighbors)
                   {
                       ref readonly var fb = ref boundaryParticles[fbIndex];
                       sumGradient += fb.Volume * kernel.ComputeGradient(f.Position, fb.Position);
                   }

                   values[fIndex] = (timeStep * timeStep) * (Vector.LengthSquared(in sumGradient) / f.Density + f.Volume * sumGradientSquared);
                });
            }

            return ComputeDiagonalElement;
        }

        #endregion
    }
}
