﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion

using System.ComponentModel;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.SystemSolver;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Solver;

namespace TamiFlu.Core.Simulation.Solver.Boundary
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("Pressure Mirroring [Akinci et al. 2012]", "1.0", "Stefan Band"), UsedImplicitly]
    internal sealed class AkinciBoundaryHandling : ABoundaryHandling, IIterativeBoundaryHandling
    {
        #region Implementation of IIterativeBoundaryHandling

        /// <inheritdoc />
        public ComputeDiagonalElements CreateComputeDiagonalElementFunc(IParticleContext particleContext)
        {
            return CreateStaticComputeDiagonalElementFunc(this, particleContext);
        }

        /// <inheritdoc />
        public ComputeSourceTerms CreateComputeSourceTermFunc(IParticleContext particleContext)
        {
            return CreateStaticComputeSourceTermFunc(this, particleContext);
        }

        /// <inheritdoc />
        public InitializeVector CreateInitializeVectorFunc(IParticleContext particleContext, float factor)
        {
            return CreateStaticInitializeVectorFunc(this, particleContext, factor);
        }

        /// <inheritdoc />
        public ComputeMatrixVectorProduct CreateComputeMatrixVectorProductFunc(IParticleContext particleContext, ref Vector[] pressureAccelerations)
        {
            return CreateStaticComputeMatrixVectorProductFunc(this, particleContext, ref pressureAccelerations);
        }

        #endregion

        #region Methods

        [NotNull]
        public static ComputeDiagonalElements CreateStaticComputeDiagonalElementFunc([NotNull] IBoundaryHandling boundaryHandling, [NotNull] IParticleContext particleContext)
        {
            void ComputeDiagonalElement(ref float[] diagonalElements)
            {
                var fluidParticles = particleContext.FluidParticles;
                var boundaryParticles = particleContext.BoundaryParticles;
                var values = EnsureLength(ref diagonalElements, fluidParticles.Length);
                var timeStep = particleContext.TimeStep.TimeStepValue;
                var kernel = particleContext.Kernel;

                // compute diagonal elements
                particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
                {
                    var sumGradient = Vector.Zero;
                    var sumGradientSquared = 0f;

                    foreach (var ffIndex in f.FluidNeighbors)
                    {
                        ref readonly var ff = ref fluidParticles[ffIndex];
                        var gradient = kernel.ComputeGradient(f.Position, ff.Position);
                        sumGradient += ff.Volume * gradient;
                        sumGradientSquared += ff.Volume / ff.Density * Vector.LengthSquared(in gradient);
                    }

                    var sumGradient2 = sumGradient;
                    foreach (var fbIndex in f.BoundaryNeighbors)
                    {
                        ref readonly var fb = ref boundaryParticles[fbIndex];
                        var gradient = kernel.ComputeGradient(f.Position, fb.Position);
                        sumGradient += fb.Volume * gradient;
                        sumGradient2 += (fb.Properties.PressureForceFactor * fb.Volume) * gradient;
                    }

                    values[fIndex] = (timeStep * timeStep) * (Vector.Dot(in sumGradient, in sumGradient2) / f.Density + f.Volume * sumGradientSquared);
                });
            }

            return ComputeDiagonalElement;
        }

        [NotNull]
        public static ComputeSourceTerms CreateStaticComputeSourceTermFunc([NotNull] IBoundaryHandling boundaryHandling, [NotNull] IParticleContext particleContext)
        {
            void ComputeSourceTerm(ref float[] sourceTerms)
            {
                var values = EnsureLength(ref sourceTerms, particleContext.FluidParticles.Length);
                var restVolume = GlobalConstants.Instance.RestVolume;
                var timeStep = particleContext.TimeStep.TimeStepValue;

                boundaryHandling.ComputeDivergence(particleContext,
                                                   (in FluidParticle f, in ParticleIndex fIndex) => f.Velocity,
                                                   (in BoundaryParticle b, in ParticleIndex bIndex) => b.Velocity,
                                                   (ref FluidParticle f, in ParticleIndex fIndex, float velocityDivergence) => values[fIndex] = (restVolume / f.Volume - 1f) - velocityDivergence * timeStep,
                                                   null);
            }

            return ComputeSourceTerm;
        }

        [NotNull]
        public static InitializeVector CreateStaticInitializeVectorFunc([NotNull] IBoundaryHandling boundaryHandling, [NotNull] IParticleContext particleContext, float factor)
        {
            void InitializeVector(ref float[] vectors)
            {
                var values = EnsureLength(ref vectors, particleContext.FluidParticles.Length);

                particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) => values[fIndex] = f.Pressure * factor);
            }

            return InitializeVector;
        }

        [NotNull]
        public static ComputeMatrixVectorProduct CreateStaticComputeMatrixVectorProductFunc([NotNull] IBoundaryHandling boundaryHandling, [NotNull] IParticleContext particleContext, [NotNull] ref Vector[] pressureAccelerations)
        {
            var fluidPressureAccelerations = EnsureLength(ref pressureAccelerations, particleContext.FluidParticles.Length);

            void ComputeMatrixVectorProduct(float[] vectors, ref float[] matrixVectorProducts)
            {
                var values = EnsureLength(ref matrixVectorProducts, particleContext.FluidParticles.Length);
                var timeStep = particleContext.TimeStep.TimeStepValue;

                // compute pressure accelerations
                var kernel = particleContext.Kernel;
                var fluidParticles = particleContext.FluidParticles;
                var boundaryParticles = particleContext.BoundaryParticles;

                particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
                {
                    var fPressure = vectors[fIndex];
                    var fPressureGradient = Vector.Zero;

                    foreach (var ffIndex in f.FluidNeighbors)
                    {
                        ref readonly var ff = ref fluidParticles[ffIndex];
                        fPressureGradient += (ff.Volume * (fPressure + vectors[ffIndex])) * kernel.ComputeGradient(f.Position, ff.Position);
                    }

                    foreach (var fbIndex in f.BoundaryNeighbors)
                    {
                        ref readonly var fb = ref boundaryParticles[fbIndex];
                        fPressureGradient += (fb.Volume * fb.Properties.PressureForceFactor * fPressure) * kernel.ComputeGradient(f.Position, fb.Position);
                    }

                    fluidPressureAccelerations[fIndex] = fPressureGradient / -f.Density;
                });

                // compute divergence of pressure acceleration
                boundaryHandling.ComputeDivergence(particleContext,
                                                   (in FluidParticle f, in ParticleIndex fIndex) => fluidPressureAccelerations[fIndex],
                                                   (in BoundaryParticle b, in ParticleIndex bIndex) => Vector.Zero,
                                                   (ref FluidParticle f, in ParticleIndex fIndex, float pressureAccelerationDivergence) => values[fIndex] = (timeStep * timeStep) * pressureAccelerationDivergence,
                                                   null);
            }

            return ComputeMatrixVectorProduct;
        }

        #endregion
    }
}
