﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion

using System;
using System.ComponentModel;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.SystemSolver;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Kernel;

namespace TamiFlu.Core.Simulation.Solver.Boundary
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("Pressure Extrapolation with MLS [Band et al. 2018]", "1.1", "Stefan Band"), UsedImplicitly]
    internal sealed class LocalPressureBoundaries : ABoundaryHandling, IIterativeBoundaryHandling
    {
        #region Implementation of IIterativeBoundaryHandling

        /// <inheritdoc />
        public ComputeDiagonalElements CreateComputeDiagonalElementFunc(IParticleContext particleContext)
        {
            return AdamiBoundaryHandling.CreateStaticComputeDiagonalElementFunc(particleContext);
        }

        /// <inheritdoc />
        public ComputeSourceTerms CreateComputeSourceTermFunc(IParticleContext particleContext)
        {
            return AkinciBoundaryHandling.CreateStaticComputeSourceTermFunc(this, particleContext);
        }

        /// <inheritdoc />
        public InitializeVector CreateInitializeVectorFunc(IParticleContext particleContext, float factor)
        {
            return AkinciBoundaryHandling.CreateStaticInitializeVectorFunc(this, particleContext, factor);
        }

        /// <inheritdoc />
        public ComputeMatrixVectorProduct CreateComputeMatrixVectorProductFunc(IParticleContext particleContext, ref Vector[] pressureAccelerations)
        {
            var fluidPressureAccelerations = EnsureLength(ref pressureAccelerations, particleContext.FluidParticles.Length);

            void ComputeMatrixVectorProduct(float[] vectors, ref float[] matrixVectorProducts)
            {
                var values = EnsureLength(ref matrixVectorProducts, particleContext.FluidParticles.Length);
                var timeStep = particleContext.TimeStep.TimeStepValue;

                // compute boundary pressures
                var boundaryPressures = new float[particleContext.BoundaryParticles.Length];
                {
                    var kernel = particleContext.Kernel;
                    var fluidParticles = particleContext.FluidParticles;

                    particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b, in ParticleIndex bIndex) =>
                                                                          boundaryPressures[bIndex] = Math.Max(0f, ExtrapolateMLSPressureForBoundaryParticle(in b, fluidParticles, kernel, vectors)));
                }

                // compute pressure accelerations
                ComputeGradient(particleContext,
                                (in FluidParticle f, in ParticleIndex fIndex) => vectors[fIndex],
                                (in BoundaryParticle b, in ParticleIndex bIndex) => boundaryPressures[bIndex],
                                (ref FluidParticle f, in ParticleIndex fIndex, in Vector pressureGradient) => fluidPressureAccelerations[fIndex] = pressureGradient / -f.Density,
                                null);

                // compute divergence of pressure acceleration
                ComputeDivergence(particleContext,
                                  (in FluidParticle f, in ParticleIndex fIndex) => fluidPressureAccelerations[fIndex],
                                  (in BoundaryParticle b, in ParticleIndex bIndex) => Vector.Zero,
                                  (ref FluidParticle f, in ParticleIndex fIndex, float pressureAccelerationDivergence) => values[fIndex] = (timeStep * timeStep) * pressureAccelerationDivergence,
                                  null);
            }

            return ComputeMatrixVectorProduct;
        }

        #endregion

        #region Methods

        private static void MatrixSafeInverse2x2(float xx, float xy, float yy, out float inverse_xx, out float inverse_xy, out float inverse_yy)
        {
            var determinant = xx * yy - xy * xy;
            if (Math.Abs(determinant) > Vector.FloatEpsilon)
            {
                inverse_xx = yy / determinant;
                inverse_xy = -xy / determinant;
                inverse_yy = xx / determinant;
            }
            else
            {
                inverse_xx = 0f;
                inverse_xy = 0f;
                inverse_yy = 0f;
            }
        }

        private static Vector MatrixMul2x2(float xx, float xy, float yy, in Vector value)
        {
            return new Vector(xx * value.X + xy * value.Y,
                              xy * value.X + yy * value.Y);
        }

        private static float ExtrapolateMLSPressureForBoundaryParticle(in BoundaryParticle b,
                                                                       [NotNull] FluidParticle[] fluidParticles,
                                                                       [NotNull] IKernel kernel,
                                                                       [NotNull] float[] fluidPressures)
        {
            var sumWeightedPosition = Vector.Zero;
            var sumWeights = 0f;

            foreach (var bfIndex in b.FluidNeighbors)
            {
                ref readonly var bf = ref fluidParticles[bfIndex];
                var kernelValue = bf.Volume * kernel.ComputeValue(b.Position, bf.Position);
                sumWeightedPosition += bf.Position * kernelValue;
                sumWeights += kernelValue;
            }

            if (sumWeights > Vector.FloatEpsilon)
            {
                var weightedPosition = sumWeightedPosition / sumWeights;

                var xx = 0f;
                var xy = 0f;
                var yy = 0f;
                var alpha = 0f;
                var sourceTerm = Vector.Zero;

                foreach (var bfIndex in b.FluidNeighbors)
                {
                    ref readonly var bf = ref fluidParticles[bfIndex];
                    var kernelValue = bf.Volume * kernel.ComputeValue(b.Position, bf.Position);
                    var r = bf.Position - weightedPosition;

                    sourceTerm += r * (kernelValue * fluidPressures[bfIndex]);
                    alpha += (kernelValue * fluidPressures[bfIndex]) / sumWeights;

                    xx += kernelValue * (r.X * r.X);
                    xy += kernelValue * (r.X * r.Y);
                    yy += kernelValue * (r.Y * r.Y);
                }

                // compute inverse of symmetric 2x2 matrix
                MatrixSafeInverse2x2(xx, xy, yy, out var inverse_xx, out var inverse_xy, out var inverse_yy);
                var beta_gamma = MatrixMul2x2(inverse_xx, inverse_xy, inverse_yy, in sourceTerm);

                return alpha + Vector.Dot(in beta_gamma, b.Position - weightedPosition);
            }

            return 0f;
        }

        #endregion
    }
}
