﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.ComponentModel;
using System.Text;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.Forces;
using TamiFlu.Core.Simulation.Solver.Boundary;
using TamiFlu.Core.Simulation.SystemSolver;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Rendering;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Forces;
using TamiFlu.Interfaces.Simulation.Solver;

namespace TamiFlu.Core.Simulation.Solver
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("IISPH", "1.0", "Stefan Band"), UsedImplicitly]
    internal sealed class IISPHFluidSolver : IFluidSolver, IDebugDrawable
    {
        #region Constants

        private const float DefaultOldPressure = 0f;

        #endregion

        #region Fields

        [CanBeNull]
        private ISystemSolverResult m_LastSolverResult;

        #endregion

        #region Properties

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IExternalForce>))]
        public IExternalForce GravityForce { get; set; } = new GravityForce();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<ISurfaceTensionForce>))]
        public ISurfaceTensionForce SurfaceTensionForce { get; set; } = new NoForce();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IViscousForce>))]
        public IViscousForce ViscousForce { get; set; } = new ViscousForce();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<ISystemSolver>))]
        public ISystemSolver Solver { get; set; } = new RelaxedJacobiSystemSolver();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IIterativeBoundaryHandling>))]
        public IIterativeBoundaryHandling BoundaryHandling { get; set; } = new AkinciBoundaryHandling();

        [Browsable(true), DefaultValue(DefaultOldPressure)]
        public float OldPressure { get; set; } = DefaultOldPressure;
        
        #endregion

        #region Implementation of IFluidSolver

        /// <inheritdoc />
        public void Simulate(IParticleContext particleContext)
        {
            var performance = particleContext.PerformanceMeasurement;
            performance.Measure("Computation time for simulation", () =>
            { 
                // (1) search neighbors
                performance.Measure("Computation time for searching neighbors", () =>
                {
                    particleContext.Neighborhood.SearchNeighbors(particleContext);
                });

                // (2) compute volumes
                var boundaryHandling = BoundaryHandling;
                performance.Measure("Computation time for volumes", () =>
                {
                    boundaryHandling.ComputeVolume(particleContext);
                });

                // (3) compute maximum timestep
                var timeStep = particleContext.TimeStep;
                timeStep.ComputeMaximumTimeStep(particleContext);
                var timeStepValue = timeStep.TimeStepValue;

                performance.Measure("Computation time for external accelerations", () =>
                {
                    // (4) set accelerations to zero
                    particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) => f.Acceleration = Vector.Zero);

                    // (5) apply external forces
                    GravityForce.ApplyToFluidParticles(particleContext);
                    ViscousForce.ApplyToFluidParticles(particleContext);
                    SurfaceTensionForce.ApplyToFluidParticles(particleContext);
                    particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) => f.Velocity += timeStepValue * f.Acceleration);
                });

                // (6) solve PPE
                var pressureAccelerations = performance.Measure("Computation time for solving the PPE", () => SolvePPE(particleContext));

                // (7) advection
                performance.Measure("Computation time for advection", () =>
                {
                    particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) =>
                    {
                        var pressureAcceleration = pressureAccelerations[fIndex];
                        f.Acceleration += pressureAcceleration;
                        f.Velocity += timeStepValue * pressureAcceleration;
                        f.Position += timeStepValue * f.Velocity;
                    });
                });
                timeStep.UpdateCurrentTime(timeStepValue);
            });
        }

        #endregion

        #region Implementation of IDebugDrawable

        /// <inheritdoc />
        public void AppendText(StringBuilder stringBuilder, IParticleContext particleContext)
        {
            var solverResult = m_LastSolverResult;
            if (!(solverResult is null))
            {
                stringBuilder.AppendLine($"PPE solver converged: {solverResult.IsConverged}");
                stringBuilder.AppendLine($"PPE iterations: {solverResult.NumberOfIterations}");
                stringBuilder.AppendLine($"Error: {solverResult.SolutionError}\u2009%");
            }
        }

        #endregion

        #region Methods
        
        [NotNull]
        private Vector[] SolvePPE([NotNull] IParticleContext particleContext)
        {
            var solver = Solver;
            var boundaryHandling = BoundaryHandling;
            solver.ComputeDiagonalElementFunc = boundaryHandling.CreateComputeDiagonalElementFunc(particleContext);
            solver.ComputeSourceTermFunc = boundaryHandling.CreateComputeSourceTermFunc(particleContext);
            solver.InitializeVectorFunc = boundaryHandling.CreateInitializeVectorFunc(particleContext, OldPressure);

            var pressureAccelerations = Array.Empty<Vector>();
            solver.ComputeMatrixVectorProductFunc = boundaryHandling.CreateComputeMatrixVectorProductFunc(particleContext, ref pressureAccelerations);

            var solverResult = solver.SolveSystem(particleContext.Parallel);
            var solutionVector = solverResult.SolutionVector;
            particleContext.ForEachFluidParticleInParallel((ref FluidParticle f, in ParticleIndex fIndex) => f.Pressure = solutionVector[fIndex]);
            m_LastSolverResult = solverResult;

            return pressureAccelerations;
        }
        
        #endregion
    }
}
