﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System.ComponentModel;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation.Forces;
using TamiFlu.Core.Simulation.Solver.Boundary;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Forces;
using TamiFlu.Interfaces.Simulation.Solver;

namespace TamiFlu.Core.Simulation.Solver
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("State Equation Solver", "1.2", "Stefan Band"), UsedImplicitly]
    internal sealed class StateEquationFluidSolver : IFluidSolver
    {
        #region Properties

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IStateEquation>))]
        public IStateEquation StateEquation { get; set; } = new StateEquation();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IExternalForce>))]
        public IExternalForce GravityForce { get; set; } = new GravityForce();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<ISurfaceTensionForce>))]
        public ISurfaceTensionForce SurfaceTensionForce { get; set; } = new NoForce();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IViscousForce>))]
        public IViscousForce ViscousForce { get; set; } = new ViscousForce();

        [NotNull, Browsable(true), TypeConverter(typeof(PluginTypeConverter<IPressureForce>))]
        public IPressureForce PressureForce { get; set; } = new PressureForce();

        [NotNull, Browsable(false)]
        public AkinciBoundaryHandling BoundaryHandling { get; set; } = new AkinciBoundaryHandling();

        #endregion

        #region Implementation of IFluidSolver

        /// <inheritdoc />
        public void Simulate(IParticleContext particleContext)
        {
            var performance = particleContext.PerformanceMeasurement;
            performance.Measure("Computation time for simulation", () =>
            {
                // (1) search neighbors
                performance.Measure("Computation time for searching neighbors", () =>
                {
                    particleContext.Neighborhood.SearchNeighbors(particleContext);
                });

                // (2) compute volumes
                var boundaryHandling = BoundaryHandling;
                performance.Measure("Computation time for volumes", () =>
                {
                    boundaryHandling.ComputeVolume(particleContext);
                });

                // (3) compute pressure of fluid particles via equation of state
                performance.Measure("Computation time for evaluating state equation", () =>
                {
                    var stateEquation = StateEquation;
                    particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) => f.Pressure = stateEquation.ComputePressure(f.Density, f.Properties.RestDensity));
                });

                performance.Measure("Computation time for forces", () =>
                {
                    // (4) set accelerations to zero
                    particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) => f.Acceleration = Vector.Zero);

                    // (5) apply forces
                    GravityForce.ApplyToFluidParticles(particleContext);
                    ViscousForce.ApplyToFluidParticles(particleContext);
                    SurfaceTensionForce.ApplyToFluidParticles(particleContext);
                    PressureForce.ApplyToFluidParticles(particleContext);
                });

                // (6) compute maximum timestep
                var timeStep = particleContext.TimeStep;
                timeStep.ComputeMaximumTimeStep(particleContext);
                var timeStepValue = timeStep.TimeStepValue;

                // (7) advection
                performance.Measure("Computation time for advection", () =>
                {
                    particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) =>
                    {
                        f.Velocity += timeStepValue * f.Acceleration;
                        f.Position += timeStepValue * f.Velocity;
                    });
                });
                timeStep.UpdateCurrentTime(timeStepValue);
            });
        }

        #endregion
    }
}
