﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.ComponentModel;
using System.Text;
using JetBrains.Annotations;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Rendering;
using TamiFlu.Interfaces.Simulation;
using TamiFlu.Interfaces.Simulation.Solver;

namespace TamiFlu.Core.Simulation.Solver
{
    [TypeConverter(typeof (ExpandableObjectConverter)), Plugin("CFL Time Step", "1.1", "Stefan Band"), UsedImplicitly]
    internal sealed class TimeStep : ITimeStep, IDebugDrawable
    {
        #region Constants

        private const float DefaultCFLLambda = 0.4f;
        private const float DefaultMinimumTimeStep = 1e-6f;
        private const float DefaultMaximumTimeStep = 1e-3f;

        #endregion

        #region Properties

        [Browsable(true), DefaultValue(DefaultCFLLambda), UsedImplicitly]
        public float CFLLambda { get; set; } = DefaultCFLLambda;

        [Browsable(true), DefaultValue(DefaultMinimumTimeStep), UsedImplicitly]
        public float MinimumTimeStep { get; set; } = DefaultMinimumTimeStep;

        [Browsable(true), DefaultValue(DefaultMaximumTimeStep), UsedImplicitly]
        public float MaximumTimeStep { get; set; } = DefaultMaximumTimeStep;

        #endregion

        #region Implementation of ITimeStep

        /// <inheritdoc />
        [Browsable(false)]
        public float TimeStepValue { get; private set; }

        /// <inheritdoc />
        [Browsable(false)]
        public float CurrentTime { get; private set; }

        /// <inheritdoc />
        public void ComputeMaximumTimeStep(IParticleContext particleContext)
        {
            var fluidMaximumVelocitySquared = particleContext.ForEachFluidParticleReduceInParallel((ref FluidParticle f) => Vector.LengthSquared(f.Velocity), Math.Max, 0f);
            if (float.IsNaN(fluidMaximumVelocitySquared))
            {
                fluidMaximumVelocitySquared = 0f;
            }

            var boundaryMaximumVelocitySquared = particleContext.ForEachBoundaryParticleReduceInParallel((ref BoundaryParticle b) => Vector.LengthSquared(b.Velocity), Math.Max, 0f);
            if (float.IsNaN(boundaryMaximumVelocitySquared))
            {
                boundaryMaximumVelocitySquared = 0f;
            }

            var maximumVelocity = (float) Math.Sqrt(Math.Max(fluidMaximumVelocitySquared, boundaryMaximumVelocitySquared));

            TimeStepValue = maximumVelocity <= float.Epsilon
                                ? MaximumTimeStep
                                : Math.Max(MinimumTimeStep, Math.Min(MaximumTimeStep, CFLLambda * GlobalConstants.Instance.Spacing / maximumVelocity));
        }

        /// <inheritdoc />
        public void ResetCurrentTime()
        {
            CurrentTime = 0f;
        }

        /// <inheritdoc />
        public void UpdateCurrentTime(float dt)
        {
            CurrentTime += dt;
        }
        
        #endregion

        #region Implementation of IDebugDrawable

        /// <inheritdoc />
        public void AppendText(StringBuilder stringBuilder, IParticleContext particleContext)
        {
            stringBuilder.AppendLine($"Time: {CurrentTime}\u2009s");
            stringBuilder.AppendLine($"Time step \u0394t: {TimeStepValue}\u2009s");
        }

        #endregion
    }
}
