﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion

using System;
using System.ComponentModel;
using JetBrains.Annotations;
using TamiFlu.Interfaces;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.Simulation.SystemSolver
{
    [TypeConverter(typeof(ExpandableObjectConverter)), Plugin("Relaxed Jacobi", "1.0", "Stefan Band"), UsedImplicitly]
    internal sealed class RelaxedJacobiSystemSolver : ISystemSolver
    {
        #region Constants

        private const float DefaultRelaxation = 0.5f;
        private const float DefaultToleratedError = 0.1f;
        private const int DefaultMinimumNumberOfIterations = 2;
        private const int DefaultMaximumNumberOfIterations = 100;
        private const bool DefaultClampValues = true;

        #endregion

        #region Nested Types: RelaxedJacobiSystemSolverResult

        private sealed class RelaxedJacobiSystemSolverResult : ISystemSolverResult
        {
            #region Implementation of ISystemSolverResult

            /// <inheritdoc />
            public int NumberOfIterations { get; set; }

            /// <inheritdoc />
            public bool IsConverged { get; set; }

            /// <inheritdoc />
            public float SolutionError { get; set; }

            /// <inheritdoc />
            public float[] SolutionVector { get; set; }

            #endregion
        }

        #endregion

        #region Properties

        [Browsable(true), DefaultValue(DefaultRelaxation)]
        public float Relaxation { get; set; } = DefaultRelaxation;

        [Browsable(true), DefaultValue(DefaultToleratedError)]
        public float ToleratedError { get; set; } = DefaultToleratedError;

        [Browsable(true), DefaultValue(DefaultClampValues)]
        public bool ClampValues { get; set; } = DefaultClampValues;

        [Browsable(true), DefaultValue(DefaultMinimumNumberOfIterations)]
        public int MinimumNumberOfIterations { get; set; } = DefaultMinimumNumberOfIterations;

        [Browsable(true), DefaultValue(DefaultMaximumNumberOfIterations)]
        public int MaximumNumberOfIterations { get; set; } = DefaultMaximumNumberOfIterations;

        #endregion

        #region Implementation of ISystemSolver

        /// <inheritdoc />
        [Browsable(false)]
        public ComputeDiagonalElements ComputeDiagonalElementFunc { get; set; }

        /// <inheritdoc />
        [Browsable(false)]
        public ComputeSourceTerms ComputeSourceTermFunc { get; set; }

        /// <inheritdoc />
        [Browsable(false)]
        public InitializeVector InitializeVectorFunc { get; set; }

        /// <inheritdoc />
        [Browsable(false)]
        public ComputeMatrixVectorProduct ComputeMatrixVectorProductFunc { get; set; }

        /// <inheritdoc />
        public ISystemSolverResult SolveSystem(IParallelContext parallelContext)
        {
            var sourceTerms = Array.Empty<float>();
            ComputeSourceTermFunc(ref sourceTerms);

            var vectors = Array.Empty<float>();
            InitializeVectorFunc(ref vectors);

            var result = new RelaxedJacobiSystemSolverResult {SolutionVector = vectors};

            var diagonalElements = Array.Empty<float>();
            ComputeDiagonalElementFunc(ref diagonalElements);

            var matrixVectorProducts = Array.Empty<float>();
            for (var iteration = 0; iteration < MaximumNumberOfIterations; ++iteration)
            {
                result.NumberOfIterations = iteration;
                ComputeMatrixVectorProductFunc(vectors, ref matrixVectorProducts);

                if (iteration < MinimumNumberOfIterations || !IsConverged())
                {
                    var relaxation = Relaxation;
                    parallelContext.ForEachRange(0, vectors.Length, (startIndex, endIndex) =>
                    {
                        for (var i = startIndex; i < endIndex; ++i)
                        {
                            float newValue;

                            var diagonalElement = diagonalElements[i];
                            // ReSharper disable once CompareOfFloatsByEqualityOperator
                            if (Math.Abs(diagonalElement) <= Vector.FloatEpsilon)
                            {
                                newValue = 0f;
                            }
                            else
                            {
                                newValue = vectors[i] + (relaxation / diagonalElement) * (sourceTerms[i] - matrixVectorProducts[i]);

                                if (ClampValues)
                                {
                                    newValue = Math.Max(0f, newValue);
                                }
                            }

                            vectors[i] = newValue;
                        }
                    });
                }
                else
                {
                    break;
                }

                bool IsConverged()
                {
                    (float Error, int Count) initValue = (0f, 0);
                    var reducedError = parallelContext.ReduceRange(0,
                                                                   vectors.Length,
                                                                   i =>
                                                                   {
                                                                       // exclude values where the diagonalelement is zero
                                                                       // ReSharper disable once CompareOfFloatsByEqualityOperator
                                                                       if (Math.Abs(diagonalElements[i]) <= Vector.FloatEpsilon)
                                                                       {
                                                                           return initValue;
                                                                       }

                                                                       // ReSharper disable once CompareOfFloatsByEqualityOperator
                                                                       var localError = ClampValues && vectors[i] == 0f ? 0f : Math.Abs(sourceTerms[i] - matrixVectorProducts[i]);
                                                                       return (Error: localError, Count: 1);
                                                                   },
                                                                   (lhs, rhs) => (lhs.Error + rhs.Error, lhs.Count + rhs.Count),
                                                                   initValue);

                    var solutionErrorPercentage = reducedError.Count != 0 ? (100f * reducedError.Error / reducedError.Count) : 0f;
                    result.SolutionError = solutionErrorPercentage;
                    var converged = solutionErrorPercentage <= ToleratedError;
                    result.IsConverged = converged;
                    return converged;
                }
            }

            return result;
        }

        #endregion
    }
}
