﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Threading;
using Eto.Forms;
using JetBrains.Annotations;
using TamiFlu.Core.UI.Rendering;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.UI.Commands
{
    internal class SimulateOneCommand : Command
    {
        #region Fields

        [NotNull] private readonly IParticleContext m_ParticleContext;
        [NotNull] private readonly RenderingControl m_RenderingControl;
        [CanBeNull] private readonly ReaderWriterLock m_SimulationLock;

        #endregion

        #region Events

        public event EventHandler BeforeSimulate;
        public event EventHandler AfterSimulate;

        #endregion

        #region Constructor

        public SimulateOneCommand([NotNull] IParticleContext particleContext, [NotNull] RenderingControl renderingControl, [CanBeNull] ReaderWriterLock simulationLock)
        {
            if (particleContext is null)
            {
                throw new ArgumentNullException(nameof(particleContext));
            }

            if (renderingControl is null)
            {
                throw new ArgumentNullException(nameof(renderingControl));
            }

            m_ParticleContext = particleContext;
            m_RenderingControl = renderingControl;
            m_SimulationLock = simulationLock;

            ToolBarText = "Step Simulation";
        }

        #endregion

        #region Methods

        private void SimulateOne()
        {
            BeforeSimulate?.Invoke(this, EventArgs.Empty);

            m_SimulationLock?.AcquireWriterLock(Timeout.Infinite);
            try
            {
                m_ParticleContext.FluidSolver.Simulate(m_ParticleContext);
                m_RenderingControl.DebugText = m_ParticleContext.GetDebugText();
            }
            finally
            {
                m_SimulationLock?.ReleaseWriterLock();
            }

            AfterSimulate?.Invoke(this, EventArgs.Empty);

            m_RenderingControl.BeginInvokeRender();
        }

        /// <inheritdoc />
        protected override void OnExecuted(EventArgs e)
        {
            base.OnExecuted(e);

            SimulateOne();
        }

        #endregion
    }
}
