﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using Eto.Drawing;
using Eto.Forms;
using JetBrains.Annotations;
using TamiFlu.Core.Simulation;
using TamiFlu.Core.UI.Commands;
using TamiFlu.Core.UI.Rendering;
using TamiFlu.Interfaces.Simulation;
using System.Threading;

namespace TamiFlu.Core.UI
{
    public sealed class MainForm : Form
    {
        #region Static Properties

        public static MainForm Instance { get; private set; }

        #endregion

        #region Fields

        [NotNull] private readonly ReaderWriterLock m_ReaderWriterLock = new ReaderWriterLock();

        #endregion

        #region Properties

        [NotNull]
        public IParticleContext ParticleContext { get; } = new ParticleContext();

        private ParticleRenderer ParticleRenderer { get; } = new ParticleRenderer();

        #endregion

        #region Constructor

        public MainForm([NotNull, ItemNotNull] params string[] args)
            : this()
        {
        }

        private MainForm()
        {
            Instance = this;

            MinimumSize = new Size(640, 480);
            Maximize();

            var renderingControl = new RenderingControl();
            renderingControl.Render += (sender, e) =>
                                       {
                                           m_ReaderWriterLock.AcquireReaderLock(Timeout.Infinite);
                                           try
                                           {
                                               var particleContext = ParticleContext;
                                               var renderingStopwatch = particleContext.PerformanceMeasurement.Restart("Rendering");
                                               particleContext.ColorComputer.ComputeColors(particleContext);

                                               ParticleRenderer.Render(particleContext.Parallel, e.Graphics, ((RenderingControl) sender).Camera, particleContext.FluidParticles);
                                               ParticleRenderer.Render(particleContext.Parallel, e.Graphics, ((RenderingControl) sender).Camera, particleContext.BoundaryParticles);
                                               renderingStopwatch.Stop();
                                           }
                                           finally
                                           {
                                               m_ReaderWriterLock.ReleaseReaderLock();
                                           }
                                       };
            renderingControl.Pick += (sender, e) =>
                                     {
                                         m_ReaderWriterLock.AcquireReaderLock(Timeout.Infinite);
                                         try
                                         {
                                             var position = e.Position;
                                             var particleContext = ParticleContext;
                                             particleContext.PickedFluidParticle = null;
                                             particleContext.PickedBoundaryParticle = null;

                                             (float DistanceSquared, ParticleIndex Index) Reduction((float, ParticleIndex) lhs, (float, ParticleIndex) rhs) => lhs.Item1 < rhs.Item1 ? lhs : rhs;
                                             (float DistanceSquared, ParticleIndex Index) FluidAction(ref FluidParticle f, in ParticleIndex fIndex) => (Vector.DistanceSquared(f.Position, in position), fIndex);
                                             (float DistanceSquared, ParticleIndex Index) BoundaryAction(ref BoundaryParticle b, in ParticleIndex bIndex) => (Vector.DistanceSquared(b.Position, in position), bIndex);
                                             var initValue = (DistanceSquared: float.MaxValue, Index: ParticleIndex.Zero);
                                             var closestFluidParticle = initValue;
                                             var closestBoundaryParticle = initValue;

                                             var neighborhood = particleContext.Neighborhood;
                                             if (neighborhood.HasNeighbors)
                                             {
                                                 neighborhood.SearchNeighbors(particleContext,
                                                                              position,
                                                                              (ref FluidParticle f) => closestFluidParticle = Reduction(closestFluidParticle, FluidAction(ref f, ParticleEquals.GetParticleIndex(ref f, particleContext.FluidParticles))),
                                                                              (ref BoundaryParticle b) => closestBoundaryParticle = Reduction(closestBoundaryParticle, BoundaryAction(ref b, ParticleEquals.GetParticleIndex(ref b, particleContext.BoundaryParticles))));
                                             }
                                             else
                                             {
                                                 // brute force search
                                                 closestFluidParticle = particleContext.ForEachFluidParticleReduceInParallel(FluidAction, Reduction, initValue);
                                                 closestBoundaryParticle = particleContext.ForEachBoundaryParticleReduceInParallel(BoundaryAction, Reduction, initValue);
                                             }

                                             var spacing = GlobalConstants.Instance.Spacing;
                                             var radiusSquared = spacing * spacing * 0.25f;
                                             if (closestBoundaryParticle.DistanceSquared < closestFluidParticle.DistanceSquared)
                                             {
                                                 if (closestBoundaryParticle.DistanceSquared < radiusSquared)
                                                 {
                                                     particleContext.PickedBoundaryParticle = closestBoundaryParticle.Index;
                                                 }
                                             }
                                             else
                                             {
                                                 if (closestFluidParticle.DistanceSquared < radiusSquared)
                                                 {
                                                     particleContext.PickedFluidParticle = closestFluidParticle.Index;
                                                 }
                                             }
                                         }
                                         finally
                                         {
                                             m_ReaderWriterLock.ReleaseReaderLock();
                                         }
                                     };

            PropertyEditorControl propertyEditor = null;
            if (Eto.Platform.Detect.IsWinForms)
            {
                propertyEditor = new PropertyEditorControl();
                propertyEditor.Handler.SelectedObject = ParticleContext;
            }

            Title = "TamiFlu";

            Content = new Splitter
                      {
                          Panel1 = renderingControl,
                          Panel2 = propertyEditor,
                          FixedPanel = SplitterFixedPanel.None,
                          RelativePosition = 0.5
                      };


            Menu = new MenuBar
                   {
                       Items =
                       {
                           new ButtonMenuItem
                           {
                               Text = "&File",
                               Items =
                               {
                                   new LoadParticleFileCommand()
                               }
                           }
                       }
                   };

            ToolBar = new ToolBar
                      {
                          Items =
                          {
                              new SimulateCommand(ParticleContext, renderingControl, m_ReaderWriterLock),
                              new SimulateOneCommand(ParticleContext, renderingControl, m_ReaderWriterLock),
                          }
                      };


            Menu.QuitItem = new Command((sender, e) => Application.Instance.Quit())
                            {
                                MenuText = "Quit",
                                Shortcut = Application.Instance.CommonModifier | Keys.Q
                            };

            var aboutItem = new ButtonMenuItem { Text = "About..." };
            aboutItem.Click += (sender, e) =>
                               {
                                   using (var dialog = new Dialog())
                                   {
                                       dialog.Content = new Label {Text = "TamiFlu."};
                                       dialog.ClientSize = new Size(200, 200);
                                       dialog.ShowModal(this);
                                   }
                               };
            Menu.AboutItem = aboutItem;
        }

        #endregion
    }
}
