﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.ComponentModel;
using System.Text;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Rendering;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.UI.Rendering.Color
{
    internal abstract class BaseColorTableComputer : IParticleColorComputer, IDebugDrawable
    {
        #region Constants

        private const bool DefaultFixedColor = false;
        private const bool DefaultGlobalColor = true;
        private const double DefaultMinimum = 0.0;
        private const double DefaultMaximum = 1.0;

        #endregion

        #region Fields

        private double m_FluidMinimum;
        private double m_FluidMaximum;
        private double m_BoundaryMinimum;
        private double m_BoundaryMaximum;
        private float m_LastTime = float.NaN;

        #endregion

        #region Properties

        [Browsable(true), DefaultValue(DefaultFixedColor), UsedImplicitly]
        public bool FixedColor { get; set; } = DefaultFixedColor;

        [Browsable(true), DefaultValue(DefaultGlobalColor), UsedImplicitly]
        public bool GlobalColor { get; set; } = DefaultGlobalColor;

        [Browsable(true), DefaultValue(DefaultMinimum), UsedImplicitly]
        public double Minimum { get; set; } = DefaultMinimum;

        [Browsable(true), DefaultValue(DefaultMaximum), UsedImplicitly]
        public double Maximum { get; set; } = DefaultMaximum;

        [Browsable(false), NotNull]
        protected abstract ARGBColorTable FluidColorTable { get; }

        [Browsable(false), NotNull]
        protected abstract ARGBColorTable BoundaryColorTable { get; }

        [Browsable(false), CanBeNull]
        protected abstract string AttributeName { get; }

        #endregion

        #region Methods

        private void UpdateMinimumAndMaximum(IParticleContext particleContext)
        {
            if (FixedColor)
            {
                m_FluidMinimum = m_BoundaryMinimum = Minimum;
                m_FluidMaximum = m_BoundaryMaximum = Maximum;
            }
            else
            {
                var currentTime = particleContext.TimeStep.CurrentTime;
                // ReSharper disable once CompareOfFloatsByEqualityOperator
                if (m_LastTime != currentTime)
                {
                    m_LastTime = currentTime;
                    Vector Reduction(Vector lhs, Vector rhs) => new Vector(Math.Min(lhs.X, rhs.X), Math.Max(lhs.Y, rhs.Y));
                    var initValue = new Vector(float.MaxValue, float.MinValue);

                    var fluidMinMax = particleContext.ForEachFluidParticleReduceInParallel((ref FluidParticle f) => new Vector(GetFluidValue(ref f)), Reduction, initValue);
                    var boundaryMinMax = particleContext.ForEachBoundaryParticleReduceInParallel((ref BoundaryParticle b) => new Vector(GetBoundaryValue(ref b)), Reduction, initValue);
                    if (GlobalColor)
                    {
                        var totalMinMax = Reduction(fluidMinMax, boundaryMinMax);
                        m_FluidMinimum = m_BoundaryMinimum = totalMinMax.X;
                        m_FluidMaximum = m_BoundaryMaximum = totalMinMax.Y;
                    }
                    else
                    {
                        m_FluidMinimum = fluidMinMax.X;
                        m_FluidMaximum = fluidMinMax.Y;
                        m_BoundaryMinimum = boundaryMinMax.X;
                        m_BoundaryMaximum = boundaryMinMax.Y;
                    }
                }
            }
        }

        #endregion

        #region Implementation of IParticleColorComputer

        /// <inheritdoc />
        public void ComputeColors(IParticleContext particleContext)
        {
            UpdateMinimumAndMaximum(particleContext);

            var fluidColorTable = FluidColorTable;
            particleContext.ForEachFluidParticleInParallel((ref FluidParticle f) => f.Color = fluidColorTable.InterpolateColor(GetFluidValue(ref f), m_FluidMinimum, m_FluidMaximum));

            var boundaryColorTable = BoundaryColorTable;
            particleContext.ForEachBoundaryParticleInParallel((ref BoundaryParticle b) => b.Color = boundaryColorTable.InterpolateColor(GetBoundaryValue(ref b), m_BoundaryMinimum, m_BoundaryMaximum));
        }

        #endregion

        #region Methods

        protected abstract float GetFluidValue(ref FluidParticle f);

        protected abstract float GetBoundaryValue(ref BoundaryParticle b);

        #endregion

        #region Implementation of IDebugDrawable

        /// <inheritdoc />
        public void AppendText(StringBuilder stringBuilder, IParticleContext particleContext)
        {
            if (!FixedColor)
            {
                var attributeName = AttributeName;
                if (!(attributeName is null))
                {
                    UpdateMinimumAndMaximum(particleContext);
                    if (GlobalColor)
                    {
                        stringBuilder.AppendLine($"{attributeName} min={m_FluidMinimum}, max={m_FluidMaximum}");
                    }
                    else
                    {
                        stringBuilder.AppendLine($"Fluid {attributeName} min={m_FluidMinimum}, max={m_FluidMaximum}");
                        stringBuilder.AppendLine($"Boundary {attributeName} min={m_BoundaryMinimum}, max={m_BoundaryMaximum}");
                    }
                }
            }
        }

        #endregion
    }
}
