﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Simulation;
using Eto.Drawing;
using TamiFlu.Interfaces;

namespace TamiFlu.Core.UI.Rendering
{
    internal sealed class ParticleRenderer
    {
        #region Constants

        private const bool RenderAsRectangle = false;
        private const bool RenderAsFill = false;
        private const float RenderAsPointSize = 4.0f;

        #endregion

        #region Methods

        private static float GetParticleSize([NotNull] Camera camera)
        {
            var projectionMatrix = camera.ProjectionMatrix;

            var clipCoordinate1 = Vector.Transform(Vector.Zero, in projectionMatrix);
            var clipCoordinate2 = Vector.Transform(new Vector(GlobalConstants.Instance.Spacing), in projectionMatrix);

            var windowCoordinate1 = camera.ToWindowCoordinates(clipCoordinate1);
            var windowCoordinate2 = camera.ToWindowCoordinates(clipCoordinate2);
            var particleSizeInWindowCoordinates = System.Numerics.Vector2.Abs(windowCoordinate2 - windowCoordinate1);

            return Math.Min(particleSizeInWindowCoordinates.X, particleSizeInWindowCoordinates.Y);
        }

        public void Render<T>([NotNull] IParallelContext parallel, [NotNull] Graphics graphics, [NotNull] Camera camera, [NotNull] T[] particles) where T : struct, IParticle
        {
            graphics.AntiAlias = true;
            graphics.ImageInterpolation = ImageInterpolation.High;
            graphics.PixelOffsetMode = PixelOffsetMode.None;

            var projectionMatrix = camera.ProjectionMatrix;
            var particleSize = GetParticleSize(camera);
            var location = camera.Location;
            var size = camera.Size;
            var bounds = graphics.ClipBounds;

            var renderInfo = new (RectangleF, Eto.Drawing.Color)[particles.Length];
            parallel.ForEachRange(0, particles.Length, (startIndex, endIndex) =>
            {
                for (var i = startIndex; i < endIndex; ++i)
                {
                    ref readonly var particle = ref particles[i];
                    var clipCoordinate = Vector.Transform(particle.Position, in projectionMatrix);
                    var windowCoordinates = Camera.ToWindowCoordinates(in clipCoordinate, in location, in size);
                    var rectanglePoint = windowCoordinates - new System.Numerics.Vector2(particleSize * 0.5f);
                    var rectangle = new RectangleF(rectanglePoint.X, rectanglePoint.Y, particleSize, particleSize);
                                          
                    Eto.Drawing.Color color;
                    // test if particle is inside the view area
                    if (bounds.Intersects(rectangle))
                    {
                        var particleColor = particle.Color;
                        color = Eto.Drawing.Color.FromArgb(particleColor.R, particleColor.G, particleColor.B, particleColor.A);
                    }
                    else
                    {
                        color = Colors.Transparent;
                    }

                    renderInfo[i] = (rectangle, color);
                }
            });

            for (var i = 0; i < renderInfo.Length; ++i)
            {
                var color = renderInfo[i].Item2;
                if (color != Colors.Transparent)
                {
                    var rectangle = renderInfo[i].Item1;

                    #pragma warning disable 162
                    if (RenderAsFill)
                    {
                        using (var brush = new SolidBrush(color))
                        {
                            if (RenderAsRectangle || particleSize < RenderAsPointSize)
                            {
                                graphics.FillRectangle(brush, rectangle);
                            }
                            else
                            {
                                graphics.FillEllipse(brush, rectangle);
                            }
                        }
                    }
                    else
                    {
                        using (var pen = new Pen(color))
                        {
                            if (RenderAsRectangle || particleSize < RenderAsPointSize)
                            {
                                graphics.DrawRectangle(pen, rectangle.X, rectangle.Y, rectangle.Width, rectangle.Height);
                            }
                            else
                            {
                                graphics.DrawEllipse(pen, rectangle);
                            }
                        }
                    }
                    #pragma warning restore 162
                }
            }
        }

        #endregion
    }
}
