﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Numerics;
using Eto.Drawing;
using Eto.Forms;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Simulation;

namespace TamiFlu.Core.UI.Rendering
{
    internal sealed class RenderingControl : Drawable
    {
        #region Fields

        private bool m_IsControlKeyDown;
        private bool m_IsMouseDown;
        private PointF m_LastMouseLocation;

        #endregion

        #region Properties

        public Camera Camera { get; } = new Camera();

        [CanBeNull]
        public string DebugText { get; set; }

        #endregion

        #region Events

        public event EventHandler<PaintEventArgs> Render;
        public event EventHandler<PickEventArgs> Pick;

        #endregion

        #region Constructor

        public RenderingControl()
            : base(false)
        {
        }

        #endregion

        #region Methods

        public void InvokeRender()
        {
            Application.Instance.Invoke(() => Update(Bounds));
        }

        public void BeginInvokeRender()
        {
            Application.Instance.Invoke(Invalidate);
        }

        /// <inheritdoc />
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);

            e.Graphics.Clear(BackgroundColor);

            Render?.Invoke(this, e);

            var text = DebugText;
            if (!string.IsNullOrEmpty(text))
            {
                e.Graphics.DrawText(SystemFonts.Message(), Colors.Black, 5f, 5f, text);
            }
        }

        /// <inheritdoc />
        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);

            Camera.Rectangle = Bounds;

            BeginInvokeRender();
        }

        /// <inheritdoc />
        protected override void OnMouseMove(MouseEventArgs e)
        {
            base.OnMouseMove(e);

            if (m_IsMouseDown && m_LastMouseLocation != e.Location)
            {
                var deltaPoint = m_LastMouseLocation - e.Location;
                m_LastMouseLocation = e.Location;

                Camera.Offset -= new Vector2(deltaPoint.X, deltaPoint.Y);

                BeginInvokeRender();
            }
        }

        /// <inheritdoc />
        protected override void OnMouseUp(MouseEventArgs e)
        {
            base.OnMouseUp(e);
            
            if (e.Buttons == MouseButtons.Primary)
            {
                if (m_IsControlKeyDown)
                {
                    var pick = Pick;
                    if (!(pick is null))
                    {
                        var location = Camera.Unproject(new Vector2(e.Location.X, e.Location.Y));
                        var pickArgs = new PickEventArgs(location);
                        pick.Invoke(this, pickArgs);
                    }
                }

                m_IsMouseDown = false;
                m_IsControlKeyDown = false;
            }
        }

        /// <inheritdoc />
        protected override void OnMouseDown(MouseEventArgs e)
        {
            base.OnMouseDown(e);
            
            if (e.Buttons == MouseButtons.Primary)
            {
                m_IsMouseDown = true;
                m_LastMouseLocation = e.Location;
                m_IsControlKeyDown = e.Modifiers == Keys.Control;
            }
        }

        /// <inheritdoc />
        protected override void OnMouseWheel(MouseEventArgs e)
        {
            base.OnMouseWheel(e);

            Camera.Zoom -= e.Delta.Height * GlobalConstants.Instance.Spacing;

            BeginInvokeRender();
        }

        #endregion
    }
}
