#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using JetBrains.Annotations;

namespace TamiFlu.Interfaces
{
    public static class AssemblyLoader
    {
        #region Static Fields

        [NotNull, ItemNotNull] private static readonly List<Assembly> Assemblies = new List<Assembly>(0);
        [NotNull, ItemNotNull] private static readonly HashSet<string> IgnoredAssemblyNames = new HashSet<string>();

        #endregion

        #region Methods

        public static void IgnoreAssemblies([NotNull, ItemNotNull] params string[] assemblyNames)
        {
            foreach (var assemblyName in assemblyNames)
            {
                IgnoredAssemblyNames.Add(assemblyName);
            }
        }

        [NotNull]
        public static IReadOnlyDictionary<string, Type> FindInterfaces([NotNull] Type type)
        {
            Load();

            var types = new Dictionary<string, Type>(0);

            foreach (var assembly in Assemblies)
            {
                foreach (var typeInAssembly in assembly.GetTypes())
                {
                    if (!typeInAssembly.IsInterface && !typeInAssembly.IsAbstract && type.IsAssignableFrom(typeInAssembly))
                    {
                        var pluginAttribute = typeInAssembly.GetCustomAttribute<PluginAttribute>(true);
                        if (!(pluginAttribute is null) && pluginAttribute.Visible && !types.ContainsKey(pluginAttribute.Name))
                        {
                            types.Add(pluginAttribute.Name, typeInAssembly);
                        }
                    }
                }
            }

            return types;
        }

        private static void Load()
        {
            if (Assemblies.Count != 0)
            {
                return;
            }

            var thisAssembly = typeof(AssemblyLoader).Assembly;
            Assemblies.Add(thisAssembly);

            var thisAssemblyLocation = thisAssembly.Location;
            foreach (var assemblyPath in new DirectoryInfo(Path.GetDirectoryName(thisAssemblyLocation)).EnumerateFiles("*.dll", SearchOption.AllDirectories))
            {
                if (IgnoredAssemblyNames.All(x => !assemblyPath.Name.Contains(x)))
                {
                    var otherAssemblyLocation = assemblyPath.FullName;
                    if (!string.Equals(thisAssemblyLocation, otherAssemblyLocation))
                    {
                        var assembly = Assembly.LoadFrom(otherAssemblyLocation);
                        if (!(assembly is null))
                        {
                            Assemblies.Add(assembly);
                        }
                    }
                }
            }
        }

        #endregion
    }
}