﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Globalization;
using System.Linq;
using System.Reflection;
using JetBrains.Annotations;

namespace TamiFlu.Interfaces
{
    public sealed class PluginTypeConverter<T> : ExpandableObjectConverter
    {
        #region Fields

        [CanBeNull] private IReadOnlyDictionary<string, Type> m_Types;
        [CanBeNull] private Dictionary<Type, T> m_Cache;

        #endregion

        #region Methods

        /// <inheritdoc />
        public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
        {
            return true; // display drop
        }

        /// <inheritdoc />
        public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
        {
            return true; // drop-down vs combo
        }

        /// <inheritdoc />
        public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
        {
            if (destinationType == typeof (string))
            {
                return true;
            }

            return base.CanConvertTo(context, destinationType);
        }

        /// <inheritdoc />
        public override object ConvertTo(ITypeDescriptorContext context, CultureInfo culture, object value, Type destinationType)
        {
            if (destinationType == typeof (string))
            {
                if (value is string stringValue)
                {
                    return stringValue;
                }

                var pluginAttribute = value.GetType().GetCustomAttribute<PluginAttribute>();
                if (!(pluginAttribute is null))
                {
                    return pluginAttribute.Name;
                }
            }

            return base.ConvertTo(context, culture, value, destinationType);
        }

        /// <inheritdoc />
        public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
        {
            if (sourceType == typeof(string))
            {
                return true;
            }

            return base.CanConvertFrom(context, sourceType);
        }

        /// <inheritdoc />
        public override object ConvertFrom(ITypeDescriptorContext context, CultureInfo culture, object value)
        {
            if (value is string stringValue && !(m_Types is null))
            {
                foreach (var kvp in m_Types)
                {
                    if (string.Equals(kvp.Key, stringValue, StringComparison.Ordinal))
                    {
                        if (m_Cache is null)
                        {
                            m_Cache = new Dictionary<Type, T>();
                        }

                        if (m_Cache.TryGetValue(kvp.Value, out var cachedValue))
                        {
                            return cachedValue;
                        }

                        var instance = (T) Activator.CreateInstance(kvp.Value);
                        m_Cache.Add(kvp.Value, instance);
                        return instance;
                    }
                }
            }

            return base.ConvertFrom(context, culture, value);
        }

        /// <inheritdoc />
        public override StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
        {
            m_Types = AssemblyLoader.FindInterfaces(typeof (T));
            return new StandardValuesCollection(m_Types.Keys.ToArray());
        }

        #endregion
    }
}
