#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Neighborhood;
using TamiFlu.Interfaces.Rendering;

namespace TamiFlu.Interfaces.Simulation
{
    public delegate void FluidParticleAction(ref FluidParticle f);
    public delegate void FluidParticleIndexAction(ref FluidParticle f, in ParticleIndex index);
    public delegate T FluidParticleFunction<out T>(ref FluidParticle f);
    public delegate T FluidParticleIndexFunction<out T>(ref FluidParticle f, in ParticleIndex index);

    public sealed class FluidParticlePhysicsProperties
    {
        #region Properties

        public float RestDensity { get; set; } = 1000f;

        public float Viscosity { get; set; } = 1e-3f;

        public float SurfaceTension { get; set; } = 0.0f;

        public float Mass => RestDensity * GlobalConstants.Instance.RestVolume;

        #endregion
    }

    public struct FluidParticle : IParticle, IEquatable<FluidParticle>
    {
        #region Properties

        public Vector Position { get; set; }

        public Vector Velocity { get; set; }

        public Vector Acceleration { get; set; }

        public float Density => Properties.Mass / Volume;

        public float Volume { get; set; }

        public float Pressure { get; set; }

        public ARGBColor Color { get; set; }

        public INeighborSet FluidNeighbors { get; set; }

        public INeighborSet BoundaryNeighbors { get; set; }

        [NotNull]
        public FluidParticlePhysicsProperties Properties { get; set; }

        [CanBeNull]
        public object UserObject { get; set; }

        #endregion

        #region Constructor

        public FluidParticle([NotNull] FluidParticlePhysicsProperties properties)
        {
            Position = Vector.Zero;
            Velocity = Vector.Zero;
            Acceleration = Vector.Zero;
            Volume = 1f;
            Pressure = 0f;
            Color = ARGBColor.Empty;
            Properties = properties;
            FluidNeighbors = null;
            BoundaryNeighbors = null;
            UserObject = null;
        }

        #endregion

        #region Methods

        /// <inheritdoc />
        public bool Equals(FluidParticle other)
        {
            return Position.Equals(other.Position) 
                   && Velocity.Equals(other.Velocity) 
                   && Acceleration.Equals(other.Acceleration) 
                   && Volume.Equals(other.Volume) 
                   && Pressure.Equals(other.Pressure) 
                   && Color.Equals(other.Color) 
                   && Equals(FluidNeighbors, other.FluidNeighbors)
                   && Equals(BoundaryNeighbors, other.BoundaryNeighbors)
                   && Properties.Equals(other.Properties)
                   && Equals(UserObject, other.UserObject);
        }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
            return !(obj is null)
                   && obj is FluidParticle
                   && Equals((FluidParticle) obj);
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            unchecked
            {
                var hashCode = Position.GetHashCode();
                hashCode = (hashCode * 397) ^ Velocity.GetHashCode();
                hashCode = (hashCode * 397) ^ Acceleration.GetHashCode();
                hashCode = (hashCode * 397) ^ Volume.GetHashCode();
                hashCode = (hashCode * 397) ^ Pressure.GetHashCode();
                hashCode = (hashCode * 397) ^ Color.GetHashCode();
                if (FluidNeighbors != null)
                {
                    hashCode = (hashCode * 397) ^ FluidNeighbors.GetHashCode();
                }
                if (BoundaryNeighbors != null)
                {
                    hashCode = (hashCode * 397) ^ BoundaryNeighbors.GetHashCode();
                }
                hashCode = (hashCode * 397) ^ Properties.GetHashCode();
                if (UserObject != null)
                {
                    hashCode = (hashCode * 397) ^ UserObject.GetHashCode();
                }
                return hashCode;
            }
        }

        #endregion
    }
}