#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion

using System;
using JetBrains.Annotations;
using TamiFlu.Interfaces.Neighborhood;
using TamiFlu.Interfaces.Rendering;

namespace TamiFlu.Interfaces.Simulation
{
    public delegate void ParticleAction<T>(ref T i) where T : struct, IParticle;
    public delegate TResult ParticleFunction<T, out TResult>(ref T i) where T : struct, IParticle;


    public interface IParticle
    {
        #region Properties

        Vector Position { get; set; }

        Vector Velocity { get; set; }

        ARGBColor Color { get; set; }

        INeighborSet FluidNeighbors { get; set; }

        INeighborSet BoundaryNeighbors { get; set; }

        #endregion
    }

    public static class ParticleEquals
    {
        #region Methods

        private static class ArrayOfTwoElements<T> where T : struct
        {
            [NotNull] public static readonly T[] Value = new T[2];
        }

        private static uint SizeOf<T>(ref T first, ref T last) where T : struct
        {
            var firstReference = __makeref(first);
            var lastReference = __makeref(last);
            unsafe
            {
                return (uint) ((byte*)*(IntPtr*)&lastReference - (byte*)*(IntPtr*)&firstReference);
            }
        }

        private static uint SizeOf<T>() where T : struct
        {
            return SizeOf(ref ArrayOfTwoElements<T>.Value[0], ref ArrayOfTwoElements<T>.Value[1]);
        }

        public static ParticleIndex GetParticleIndex<T>(ref T value, [NotNull] T[] values) where T : struct, IParticle
        {
            return new ParticleIndex((int) (SizeOf(ref values[0], ref value) / SizeOf<T>()));
        }

        public static bool ReferenceEquals(ref FluidParticle lhs, ref BoundaryParticle rhs)
        {
            return false;
        }

        public static bool ReferenceEquals(ref BoundaryParticle lhs, ref FluidParticle rhs)
        {
            return false;
        }

        public static bool ReferenceEquals<T1, T2>(ref T1 lhs, ref T2 rhs, ParticleIndex lhsIndex, ParticleIndex rhsIndex)
            where T1 : struct, IParticle
            where T2 : struct, IParticle
        {
            return typeof(T1) == typeof(T2)
                   && lhsIndex == rhsIndex
                   && ReferenceEquals(ref lhs, ref rhs);
        }

        public static bool ReferenceEquals<T1, T2>(ref T1 lhs, ref T2 rhs)
            where T1 : struct, IParticle
            where T2 : struct, IParticle
        {
            if (typeof(T1) != typeof(T2))
            {
                return false;
            }

            var lhsReference = __makeref(lhs);
            var rhsReference = __makeref(rhs);
            unsafe
            {
                var lhsPointer = (IntPtr*) &lhsReference;
                var rhsPointer = (IntPtr*) &rhsReference;
                return *lhsPointer == *rhsPointer;
            }
        }

        #endregion
    }
}