﻿#region Copyright
// Copyright (c) 2017 Stefan Band <bands@informatik.uni-freiburg.de>
//  
// Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, 
// provided that the above copyright notice and this permission notice appear in all copies.
// 
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
// INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
// FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
// LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#endregion
using System;
using System.Runtime.CompilerServices;

namespace TamiFlu.Interfaces.Simulation
{
    public struct Vector : IEquatable<Vector>
    {
        #region Constants

        public const int Dimensions = 2;
        public const float FloatEpsilon = 1e-10f;

        #endregion

        #region Fields

        public readonly float X;
        public readonly float Y;

        #endregion

        #region Properties

        public static readonly Vector Zero = new Vector(0f);

        public float this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0:
                        return X;

                    case 1:
                        return Y;

                    default:
                        throw new IndexOutOfRangeException();
                }
            }
        }

        #endregion

        #region Constructor

        /// <summary>
        /// Broadcasts value to all elements of the vector
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public Vector(float value)
            : this(value, value)
        {
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public Vector(float x, float y)
        {
            X = x;
            Y = y;
        }

        #endregion

        #region Methods

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator +(Vector lhs, Vector rhs)
        {
            return new Vector(lhs.X + rhs.X, lhs.Y + rhs.Y);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator -(Vector lhs, Vector rhs)
        {
            return new Vector(lhs.X - rhs.X, lhs.Y - rhs.Y);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator *(Vector lhs, Vector rhs)
        {
            return new Vector(lhs.X * rhs.X, lhs.Y * rhs.Y);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator *(Vector lhs, float rhs)
        {
            return new Vector(lhs.X * rhs, lhs.Y * rhs);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator *(float lhs, in Vector rhs)
        {
            return rhs * lhs;
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator /(Vector lhs, in Vector rhs)
        {
            return new Vector(lhs.X / rhs.X, lhs.Y / rhs.Y);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator /(Vector lhs, float rhs)
        {
            return new Vector(lhs.X / rhs, lhs.Y / rhs);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector operator -(Vector value)
        {
            return new Vector(-value.X, -value.Y);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool operator ==(Vector lhs, Vector rhs)
        {
            return lhs.Equals(rhs);
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static bool operator !=(Vector lhs, Vector rhs)
        {
            return !(lhs == rhs);
        }

        /// <inheritdoc />
        public bool Equals(Vector other)
        {
            return X.Equals(other.X) && Y.Equals(other.Y);
        }

        /// <inheritdoc />
        public override bool Equals(object obj)
        {
            return obj is Vector && Equals((Vector) obj);
        }

        /// <inheritdoc />
        public override int GetHashCode()
        {
            return X.GetHashCode() ^ Y.GetHashCode();
        }

        /// <summary>
        /// Returns the length of the vector.
        /// </summary>
        /// <returns>The vector's length.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static float Length(in Vector value)
        {
            return (float) Math.Sqrt(LengthSquared(in value));
        }

        /// <summary>
        /// Returns the length of the vector squared. This operation is cheaper than Length().
        /// </summary>
        /// <returns>The vector's length squared.</returns>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static float LengthSquared(in Vector value)
        {
            return Dot(in value, in value);
        }

        /// <summary>
        /// Returns the Euclidean distance between the two given points.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static float Distance(in Vector lhs, in Vector rhs)
        {
            return Length(lhs - rhs);
        }

        /// <summary>
        /// Returns the Euclidean distance squared between the two given points.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static float DistanceSquared(in Vector lhs, in Vector rhs)
        {
            return LengthSquared(lhs - rhs);
        }

        /// <summary>
        /// Returns the dot product of two vectors.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static float Dot(in Vector lhs, in Vector rhs)
        {
            return (lhs.X * rhs.X) + (lhs.Y * rhs.Y);
        }

        /// <summary>
        /// Returns a vector whose elements are the minimum of each of the pairs of elements in the two source vectors.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Min(in Vector lhs, in Vector rhs)
        {
            return new Vector(Math.Min(lhs.X, rhs.X), Math.Min(lhs.Y, rhs.Y));
        }

        /// <summary>
        /// Returns a vector whose elements are the maximum of each of the pairs of elements in the two source vectors
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Max(in Vector lhs, in Vector rhs)
        {
            return new Vector(Math.Max(lhs.X, rhs.X), Math.Max(lhs.Y, rhs.Y));
        }

        /// <summary>
        /// Returns a vector whose elements are the absolute values of each of the source vector's elements.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Abs(in Vector value)
        {
            return new Vector(Math.Abs(value.X), Math.Abs(value.Y));
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Floor(in Vector value)
        {
            return new Vector((float) Math.Floor(value.X), (float) Math.Floor(value.Y));
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Ceiling(in Vector value)
        {
            return new Vector((float) Math.Ceiling(value.X), (float) Math.Ceiling(value.Y));
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Truncate(in Vector value)
        {
            return new Vector((float) Math.Truncate(value.X), (float) Math.Truncate(value.Y));
        }

        /// <summary>
        /// Returns a vector whose elements are the square root of each of the source vector's elements.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Sqrt(in Vector value)
        {
            return new Vector((float) Math.Sqrt(value.X), (float) Math.Sqrt(value.Y));
        }

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Transform(in Vector value, in System.Numerics.Matrix4x4 matrix)
        {
            return new Vector((value.X * matrix.M11 + value.Y * matrix.M21) + matrix.M41,
                              (value.X * matrix.M12 + value.Y * matrix.M22) + matrix.M42);
        }

        /// <summary>
        /// Returns a vector with the same direction as the given vector, but with a length of 1.
        /// </summary>
        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public static Vector Normalize(in Vector value)
        {
            return value / Length(in value);
        }

        /// <inheritdoc />
        public override string ToString()
        {
            return $"X={X} Y={Y}";
        }

        #endregion
    }
}
